# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

"""
This script is called from `rai_toolbox.mushin.lightning.launchers.HydraDDP
"""

import logging
from typing import Optional

import hydra
from pytorch_lightning import LightningDataModule, LightningModule, Trainer

from ..hydra import zen

log = logging.getLogger(__name__)


def task(
    trainer: Trainer,
    module: LightningModule,
    datamodule: Optional[LightningDataModule] = None,
    pl_testing: bool = False,
    pl_local_rank: int = 0,
) -> None:
    if pl_testing:
        log.info(f"Rank {pl_local_rank}: Launched subprocess using Training.test")
        trainer.test(module, datamodule=datamodule)
    else:
        log.info(f"Rank {pl_local_rank}: Launched subprocess using Training.fit")
        trainer.fit(module, datamodule=datamodule)


@hydra.main(config_path=None, config_name="config")
def main(cfg):
    zen(task)(cfg)


if __name__ == "__main__":  # pragma: no cover
    main()



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

"""
An implementation of the AugMix augmentation method specified in:

    Hendrycks, Dan, et al. "Augmix: A simple data processing method to improve
    robustness and uncertainty." arXiv preprint arXiv:1912.02781 (2019).

with reference implementation from https://github.com/google-research/augmix
"""
from numbers import Real
from typing import Callable, Optional, Sequence, Tuple, TypeVar, Union

import numpy as np

from rai_toolbox._typing import ArrayLike

_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2", bound=ArrayLike)

C1 = Callable[[_T1], _T2]
C2 = Sequence[Callable[[_T1], _T1]]


def augment_and_mix(
    datum: _T1,
    *,
    process_fn: Callable[[_T1], _T2],
    augmentations: Sequence[Callable[[_T1], _T1]],
    num_aug_chains: int = 3,
    aug_chain_depth: Union[int, Tuple[int, int]] = (1, 4),
    beta_params: Union[float, Tuple[float, float]] = (1.0, 1.0),
    dirichlet_params: Union[float, Sequence[float]] = 1.0,
    augmentation_choice_probs: Optional[Sequence[float]] = None,
) -> _T2:
    """
    Augments `datum` using a mixture of randomly-composed augmentations via a method
    called "AugMix" [1]_.

    Parameters
    ----------
    datum : T1
       The datum to be augmixed.

    num_aug_chains : int
        The number of independent "augmentation chains" that are used to produce the augmixed
        result.

    aug_chain_depth : Union[int, Tuple[int, int]]
        Determines that number of augmentations that are randomly sampled (with replacement) and
        composed to form each augmentation chain.

        If a tuple of values are provided, these are used as the lower (inclusive) and
        upper (exclusive) bounds on a uniform integer-valued distribution from which the depth
        will be sampled for each augmentation chain.

        E.g.,

        - `mixture_depth=2` means that each augmentation chain will consist of two (randomly sampled) augmentations composed together.
        - `aug_chain_depth=(1, 4)` means depth of any given augmentation chain is uniformly sampled from [1, 4).

    process_fn : Callable[[T1], T2]
        The preprocessing function applied to the both clean and augmixed datums before
        they are combined. The return type of `process_fn` determines the return type
        of `augment_and_mix`.

    augmentations : Sequence[Callable[[T1], T1]]
        The collection of datum augmentations that is sampled from (with replacement) to form each augmentation
        chain.

    beta_params : Tuple[float, float]
        The Beta distribution parameters to draw `m`, which weights that convex combination::

                (1 - m) * img_process_fn(datum) + m * img_process_fn(augment(datum))

        If a single value is specified, it is used as both parameters for the distribution.

    dirichlet_params : Union[float, Sequence[float]]
        The Dirichlet distribution parameters used to weight the `mixture_width` number of augmentation chains.
        If a sequence is provided, its length must match `num_aug_chains`.

    augmentation_choice_probs : Optional[Sequence[float]]
        The probabilities associated with sampling each respective entry in `augmentations`.
        If not specified, a uniform distribution over all entries of `augmentation`.

    Returns
    -------
    augmixed: T2
        The augmixed datum.

    Notes
    -----
    The following depicts AugMix with N augmentation chains. Each `augchain(...)` consists of
    composed augmentations, where the composition depth is determined by `mixture_depth`::

        (1 - m) * process_fn(img) + m * (w1 * (process_fn ∘ augchain1)(img) + ... + wN * (process_fn ∘ augchainN)(img))

    with

    - m ~ Beta
    - [w1, ..., wN] ~ Dirichlet

    Random values are drawn via NumPy's global random number generator. Thus `numpy.random.seed`
    must be set in order to obtain reproducible results. Note that, until
    PyTorch 1.9.0, there was an issue with using NumPy's global RNG in conjunction with
    DataLoaders that used multiple workers, where identical seeds were being used
    across workers and the same seed was being set at the outset of each epoch.

    https://github.com/pytorch/pytorch/pull/56488

    References
    ----------
    .. [1] Hendrycks, Dan, et al. "Augmix: A simple data processing method to improve robustness and uncertainty." arXiv preprint arXiv:1912.02781 (2019).
    """
    # See: https://numpy.org/doc/stable/reference/random/index.html?#module-numpy.random
    # for details about new rng methods in numpy

    if isinstance(beta_params, Real):
        beta_params = (float(beta_params), float(beta_params))

    if isinstance(dirichlet_params, (Real, float)):
        dirichlet_params = [float(dirichlet_params)] * num_aug_chains
    elif len(dirichlet_params) != num_aug_chains:
        raise ValueError(
            f"The number of specified dirichlet parameters ({len(dirichlet_params)}) "
            f"does not match the mixture width ({num_aug_chains})"
        )

    if num_aug_chains:
        mix_weight = float(np.random.beta(*beta_params))
    else:
        # no augmentations occur
        mix_weight = 0.0

    if augmentation_choice_probs is not None:
        augmentation_choice_probs = np.array(augmentation_choice_probs, dtype=float)
        augmentation_choice_probs /= augmentation_choice_probs.sum()

    width_weights = np.asarray(np.random.dirichlet(dirichlet_params), dtype=np.float32)

    # mixture_depths[i]: number of augmentations applied to branch-i
    # of mixture graph
    if isinstance(aug_chain_depth, Sequence):
        mixture_depths: np.ndarray = np.random.randint(
            *aug_chain_depth, size=num_aug_chains
        )
    else:
        mixture_depths = np.full(width_weights.shape, aug_chain_depth)

    out = (1.0 - mix_weight) * process_fn(datum)

    for depth, weight in zip(mixture_depths, width_weights):
        # augmentations must not mutate original datum
        augmented_datum = datum
        for op in np.random.choice(
            augmentations, size=depth, replace=True, p=augmentation_choice_probs
        ):
            op: Callable[[_T1], _T1]
            augmented_datum = op(augmented_datum)
        out += (mix_weight * weight) * process_fn(augmented_datum)
    out: _T2
    return out



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from numbers import Number
from typing import List, Optional, Tuple, TypeVar, Union, cast

import numpy as np
import torch as tr
from PIL.Image import Image
from torch import Tensor
from torchvision.transforms import ToPILImage, ToTensor

from ._fourier_basis import generate_fourier_bases

_T = TypeVar("_T", Tensor, Image)

_to_pil = ToPILImage()
_to_tensor = ToTensor()


class FourierPerturbation(tr.nn.Module):
    """
    Augments a shape-(C, H, W) image with a randomly-sampled Fourier-basis. The generated bases
    have values on [0.0. 1.0) and, by default, have a norm of 1.

    See A Fourier Perspective on Model Robustness in Computer Vision [1]_ for a detailed
    description of this augmentation and its implications.

    Notes
    -----
    All sources of randomness are derived from NumPy's global RNG. If you are using ``PyTorch < 1.9.0`` and
    ``num_workers > 0``, you must manually set NumPy's global seed in ``worker_init_fn``.

    For large images, this transform can generate many distinct bases and thus consume large amounts of memory.

    References
    ----------
    .. [1] https://arxiv.org/abs/1906.08988"""

    # number of random indices to be drawn at a given time
    _cache_size: int = 10_000

    def sample_basis(self) -> Tensor:
        """Returns shape-(H, W) basis tensor"""
        if not self._rand_index_cache:
            # It is slow to call into random.choice(..., p=...) individually.
            # Calling it once to draw many indices avoids overhead.
            self._rand_index_cache = np.random.choice(
                len(self.bases),
                size=self._cache_size,
                replace=True,
                p=self._sample_probs,
            ).tolist()

        index = self._rand_index_cache.pop()

        if isinstance(self._norm_bnds, tuple):
            norm = np.random.uniform(*self._norm_bnds, size=1).item()
            return self.bases[index] * norm
        else:
            # norm was already applied to all bases in __init__
            return self.bases[index]

    def __init__(
        self,
        size: Union[int, Tuple[int, int]],
        *,
        norm_scale: Union[float, Tuple[float, float]] = 1.0,
        dtype: Union[str, tr.dtype] = "float32",
        rand_flip_per_channel: bool = False,
        freq_bounds: Tuple[Optional[float], Optional[float]] = (0.0, None),
        num_distinct_phases: int = 1,
    ):
        """
        Parameters
        ----------
        size : Union[int, Tuple[int, int]]
            The spatial-extent of the image that the basis will augment. E.g. for
            cifar10, this would be ``32`` or ``(32, 32)``.

        norm_scale : Union[float, Tuple[float, float]] (optional, default=1.0)
            The scaling factor applied to each basis, prior to augmenting the image.
            A tuple can be supplied to indicate the ``[low, high)`` bounds of a uniform
            distribution from which the norm will be drawn for each augmentation.

            By default, each basis has a norm of 1.

        dtype : Union[str, tr.dtype] (optional, default='float32')

        rand_flip_per_channel : bool (optional, default=False)
            If ``True`` the perturbation is applied to each image-channel with a randomly drawn sign.
            E.g. the basis may be subtracted from the image's red channel and added to its blue and
            green channels.

        freq_bounds : Tuple[Optional[float], Optional[float]]
            Specifies the range of frequencies – number of oscillations over the image's extent – that will
            be represented among the bases, as ``[low, high]``. E.g. specifying ``(1, 3)`` will only augment
            images with Fourier bases that oscillate between (inclusive) 1 to 3 times. Bounds of ``None``
            default to the lowest (excluding zero) and highest possible frequencies that can be sustained
            by the given image.

        num_distinct_phases : int (optional, default=1)
            Specifies the number of phase-shifts that will be included in generating each basis. E.g. specifying
            ``num_distinct_phases=3`` will increase the total number of bases that can be sampled from by 3x;
            this is achieved by using the phase-shifts of 0, pi/3, and 2*pi/3 to diversify the bases.
        """
        super().__init__()
        if isinstance(dtype, tr.dtype):
            dtype = str(dtype).split(".")[-1]

        self._rand_index_cache: List[int] = []
        self._low_norm = None
        self._hi_norm = None

        self.last_position: Optional[Tuple[int, int]] = None

        assert num_distinct_phases > 0
        assert isinstance(rand_flip_per_channel, bool)
        self.rand_flip_per_channel = rand_flip_per_channel

        if isinstance(size, (int, Number)):
            size = (int(size), int(size))

        if not isinstance(norm_scale, (float, Number)):
            lo, hi = norm_scale
            assert 0 <= lo <= hi, norm_scale
            self._norm_bnds = (lo, hi)
        else:
            self._norm_bnds = norm_scale

        lower, upper = freq_bounds

        if lower is None:
            lower = 0.0
        if upper is None:
            upper = np.inf

        # freq=0 is excluded
        lower = max(1e-6, lower)

        if upper < lower or lower < 0:
            raise ValueError(
                f"radii-bounds must provide ordered bounds within [0, inf), got {(lower, upper)}"
            )
        self.radii_bounds = (lower, upper)

        self.bases = []
        _positions = []
        for item in generate_fourier_bases(
            nrows=size[0], ncols=size[1], dtype=dtype, factor_2pi_phase_shift=0
        ):
            _positions.append(item.position)
            self.bases.append(item.basis)

        # fourier bases (without any phase shift)
        self.bases = np.array(self.bases)  # shape-(N, image-H, image-W)
        # Fourier-space positions associated with bases
        _positions = np.array(_positions)  # used to draw random key

        # filter bases that fall outside of annulus of frequencies
        _zero_centered = _positions - np.array(size[0]) / 2
        _radii = np.linalg.norm(_zero_centered, axis=1)
        in_bounds = np.logical_and(lower <= _radii, _radii <= upper)
        _radii = _radii[in_bounds]
        _positions = _positions[in_bounds]
        self.bases = self.bases[in_bounds]

        if not _radii.size:
            raise ValueError(
                f"The specified radii-bounds, {freq_bounds}, do not include any Fourier bases"
            )

        # Populate bases with additional phase-shifted versions.
        # Phase shiftes are sampled evenly on [0, 2*pi) according to num_distinct_phases
        self.bases = [self.bases]

        for n in range(1, num_distinct_phases):
            fill = np.zeros_like(self.bases[0])

            frac_2_pi = n / num_distinct_phases

            for idx, item in enumerate(
                generate_fourier_bases(
                    nrows=size[0],
                    ncols=size[1],
                    dtype=dtype,
                    row_col_coords=_positions,
                    factor_2pi_phase_shift=frac_2_pi,
                )
            ):
                fill[idx] = item.basis
            self.bases.append(fill)

        self.bases = tr.tensor(np.vstack(self.bases))

        if isinstance(self._norm_bnds, (float, Number)):
            # norm is constant; scale bases upfront
            self.bases = self._norm_bnds * self.bases

        # Sample probabilities are adjusted so that sampling occurs evenly
        # over basis-frequencies.
        self._sample_probs = np.concatenate([1 / _radii] * num_distinct_phases)
        self._sample_probs /= self._sample_probs.sum()

    def forward(self, img: _T) -> _T:
        """
        Parameters
        ----------
        img: Union[Image, Tensor], shape-(C, H, W)
            Tensors are expected to have shape-(C, H, W) and to be on the domain [0, 1].
            The post-perturbation tensor is clipped to [0, 1]. PIL images are scaled
            and transposed appropriately (255 -> 1) prior to augmentation.

        Returns
        -------
        perturbed_img: Union[Image, Tensor], shape-(C, H, W)
        """

        _was_pil = False
        if isinstance(img, Image):
            _was_pil = True
            img = _to_tensor(img)
        elif not isinstance(img, Tensor):
            raise TypeError(
                f"`FourierPerturbation` only supports PIL images and PyTorch tensors, got: {type(img)}"
            )

        img: Tensor
        # shape-(1, H, W)
        basis = self.sample_basis().type_as(img)[None]

        if self.rand_flip_per_channel:
            channel_flip = np.random.randint(0, 2, (img.shape[0],))
            channel_flip[channel_flip == 0] = -1
            basis = basis * channel_flip.reshape(-1, 1, 1)

        img = img + basis

        if _was_pil:
            img = tr.clip_(img, min=0.0, max=1.0)
            return cast(_T, _to_pil(img))

        return cast(_T, img)

    def __repr__(self) -> str:

        return (
            self.__class__.__name__
            + f"(norm_scale={self._norm_bnds}, rand_flip_per_channel={self.rand_flip_per_channel}, radii_bounds={self.radii_bounds})"
        )



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from collections import defaultdict
from pathlib import Path
from typing import Union

import torch
from pytorch_lightning import Callback

from rai_toolbox._utils import value_check


class MetricsCallback(Callback):
    """Saves validation and test metrics stored in `trainer.callback_metrics`.

    Parameters
    ----------
    save_dir : str, optional (default=".")

    filename : str, optional (default="metrics.pt")
        The base filename used to store metrics.  For `FITTING` the file is prepended
        with "fit_" and and for `TESTING` the file is prepended with `test_`.

    Notes
    -----
    No metrics will be saved during `FITTING` if no validation metrics are calculated.
    This is a limitation of PyTorch Lightning. Future versions will save the training
    step metrics when no validation metrics are calculated.

    Examples
    --------

    >>> from pytorch_lightning import Trainer
    >>> from rai_toolbox.mushin import MetricsCallback

    >>> metrics_callback = MetricsCallback()
    >>> trainer = Trainer(callbacks=[metrics_callback])
    """

    def __init__(
        self,
        save_dir: Union[Path, str] = ".",
        filename: Union[Path, str] = "metrics.pt",
    ):
        super().__init__()
        self.save_dir = Path(save_dir)
        self.filename = value_check("filename", filename, type_=(str, Path))
        self.train_metrics = defaultdict(list)
        self.val_metrics = defaultdict(list)
        self.test_metrics = defaultdict(list)

    def _get_filename(self, stage: str):
        return self.save_dir / f"{stage}_{self.filename}"

    def _process_metrics(self, stored_metrics, metrics):
        for k, v in metrics.items():
            if isinstance(v, torch.Tensor):
                if v.ndim == 0:
                    v = v.item()
                else:
                    v = v.cpu().numpy()
            stored_metrics[k].append(v)

    def on_validation_end(self, trainer, pl_module):
        # Make sure PL is not doing it's sanity check run
        if trainer.sanity_checking:
            return self.val_metrics

        metrics = trainer.callback_metrics
        self.val_metrics["epoch"].append(pl_module.current_epoch)
        self._process_metrics(self.val_metrics, metrics)
        torch.save(self.val_metrics, self._get_filename("fit"))
        return self.val_metrics

    def on_test_end(self, trainer, pl_module):
        metrics = trainer.callback_metrics
        self._process_metrics(self.test_metrics, metrics)
        torch.save(self.test_metrics, self._get_filename("test"))
        return self.test_metrics



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from functools import wraps
from typing import Any, Callable, TypeVar, cast

from typing_extensions import Protocol


class Negateable(Protocol):
    def __neg__(self) -> Any:  # pragma: no cover
        ...


T = TypeVar("T", bound=Callable[..., Negateable])


def negate(func: T) -> T:
    """A wrapper that negates (applies the `-` operator) to the function's output.

    Parameters
    ----------
    func : Callable[..., Negateable]

    Examples
    --------
    >>> from rai_toolbox import negate
    >>> f = negate(lambda x: 2 * x)
    >>> f(1)
    -2
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        return -func(*args, **kwargs)

    return cast(T, wrapper)



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from ._fourier_basis import FourierBasis, generate_fourier_bases
from ._implementations import HeatMapEntry, create_heatmaps, normalize, perturb_batch

__all__ = [
    "generate_fourier_bases",
    "FourierBasis",
    "create_heatmaps",
    "perturb_batch",
    "normalize",
    "HeatMapEntry",
]



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from abc import abstractmethod
from typing import Any, Callable, Iterator, Optional, Sequence, Union

import torch as tr
from torch import Tensor, nn
from torch.nn.parameter import Parameter
from typing_extensions import Protocol, runtime_checkable


@runtime_checkable
class PerturbationModel(Protocol):
    """Protocol for Perturbation Models."""

    def __init__(self, *args, **kwargs) -> None:  # pragma: no cover
        ...

    @abstractmethod
    def __call__(self, data: Tensor) -> Tensor:  # pragma: no cover
        """A perturbation model should take data as the input and output the
        peturbed data.

        Parameters
        ----------
        data: Tensor
            The data to perturb.

        Returns
        -------
        Tensor
            The perturbed data with the same shape as `data`.
        """

    def parameters(
        self, recurse: bool = True
    ) -> Iterator[Parameter]:  # pragma: no cover
        ...


class AdditivePerturbation(nn.Module, PerturbationModel):
    r"""Modifies a piece or batch of data by adding a perturbation: :math:`x \rightarrow x+\delta`.



    Attributes
    ----------
    delta : torch.Tensor
        :math:`\delta` is the sole trainable parameter of `AdditivePerturbation`.
    """

    def __init__(
        self,
        data_or_shape: Union[Tensor, Sequence[int]],
        init_fn: Optional[Callable[[Tensor], None]] = None,
        *,
        device: Optional[tr.device] = None,
        dtype: Optional[tr.dtype] = None,
        delta_ndim: Optional[int] = None,
        **init_fn_kwargs: Any,
    ) -> None:
        """The init function should support data as the argument to initialize
        perturbations.

        Parameters
        ----------
        data_or_shape: Union[Tensor, Tuple[int, ...]]
            Determines the shape of the perturbation. If a tensor is supplied, its dtype
            and device are mirrored by the initialized perturbation.

            This parameter can be modified to control whether the perturbation adds
            elementwise or broadcast-adds over `x`.

        init_fn: Optional[Callable[[Tensor], None]]
            Operates in-place on a zero'd perturbation tensor to determine the
            final initialization of the perturbation.

        device: Optional[tr.device]
            If specified, takes precedent over the device associated with `data_or_shape`.

        dtype: Optional[tr.dtype] = None
            If specified, takes precedent over the dtype associated with `data_or_shape`.

        delta_ndim: Optional[int] = None
            If a positive number, determines the dimensionality of the perturbation.
            If a negative number, indicates the 'offset' from the dimensionality of
            `data_or_shape`. E.g., if `data_or_shape` has a shape (N, C, H, W), and
            if `delta_ndim=-1`, then the perturbation will have shape (C, H, W),
            and will be applied in a broadcast fashion.

        **init_fn_kwargs: Any
            Keyword arguments passed to `init_fn`.

        Examples
        --------
        **Basic Additive Perturbations**

        Let's imagine we have a batch of three shape-`(2, 2)` images (our toy data will
        be all ones) that we want to perturb. We'll randomly initialize a shape-`(3, 2,
        2)` tensor of perturbations to apply additively to the shape-`(3, 2, 2)` batch.

        >>> import torch as tr
        >>> from rai_toolbox.perturbations import AdditivePerturbation, uniform_like_l1_n_ball_
        >>> data = tr.ones(3, 2, 2)

        We provide a `generator` argument to control the RNG in `~rai_toolbox.
        perturbations.uniform_like_l1_n_ball_`.

        >>> pert_model = AdditivePerturbation(
        ...     data_or_shape=data,
        ...     init_fn=uniform_like_l1_n_ball_,
        ...     generator=tr.Generator().manual_seed(0),  # controls RNG of init
        ... )

        Accessing the initialized perturbations.

        >>> pert_model.delta
        Parameter containing:
        tensor([[[0.0885, 0.0436],
                [0.3642, 0.2720]],
                .
                [[0.3074, 0.1827],
                [0.1440, 0.2624]],
                .
                [[0.3489, 0.0528],
                [0.0539, 0.1767]]], requires_grad=True)

        Applying the perturbations to a batch of data.

        >>> pert_data = pert_model(data)
        >>> pert_data
        tensor([[[1.0885, 1.0436],
                [1.3642, 1.2720]],
                .
                [[1.3074, 1.1827],
                [1.1440, 1.2624]],
                .
                [[1.3489, 1.0528],
                [1.0539, 1.1767]]], grad_fn=<AddBackward0>)

        Involving the perturbed data in a computational graph where auto-diff is
        performed, then gradients are computed for the perturbations.

        >>> (pert_data ** 2).sum().backward()
        >>> pert_model.delta.grad
        tensor([[[2.1770, 2.0871],
                [2.7285, 2.5439]],
                .
                [[2.6148, 2.3653],
                [2.2880, 2.5247]],
                .
                [[2.6978, 2.1056],
                [2.1078, 2.3534]]])

        **Broadcasted ("Universal") Perturbations**

        Suppose that we want to use a single shape-`(2, 2)` tensor to perturb each datum
        in a batch. We can create a perturbation model in a similar manner, but
        specifying `delta_ndim=-1` indicates that our perturbation should have one
        fewer dimension than our data; whereas our batch has shape-`(N, 2, 2)`, our
        perturbation model's parameter will have shape-`(2, 2)`

        >>> pert_model = AdditivePerturbation(
        ...     data_or_shape=data,
        ...     delta_ndim=-1,
        ...     init_fn=uniform_like_l1_n_ball_,
        ...     generator=tr.Generator().manual_seed(1),  # controls RNG of init
        ... )
        >>> pert_model.delta
        Parameter containing:
        tensor([[0.2793, 0.4783],
                [0.4031, 0.3316]], requires_grad=True)

        Perturbing a batch of data now performs broadcast-addition of this tensor over
        the batch.

        >>> pert_data = pert_model(data)
        >>> pert_data
        tensor([[[1.2793, 1.4783],
                [1.4031, 1.3316]],
                .
                [[1.2793, 1.4783],
                [1.4031, 1.3316]],
                .
                [[1.2793, 1.4783],
                [1.4031, 1.3316]]], grad_fn=<AddBackward0>)

        .. important::

           Downstream reductions of this broadcast-pertubed data should involve
           a mean – not a sum – over the batch dimension so that the resulting gradient
           computed for the perturbation is not scaled by batch-size.

           >>> (pert_data ** 2).mean().backward()
           >>> pert_model.delta.grad
           tensor([[0.6397, 0.7392],
                   [0.7015, 0.6658]])

           Similarly, when using a `~rai_toolbox.optim.ParamTransformingOptimizer` to
           optimize this broadcasted perturbation, we should specify `param_ndim=None`
           to ensure that the parameter transformations are not broadcasted over our
           perturbation tensor and/or its gradient, as it has no batch dimension.
        """
        super().__init__()

        _init_kwargs = {}

        if isinstance(data_or_shape, tr.Tensor):
            shape = data_or_shape.shape
            _init_kwargs.update(
                {
                    "dtype": data_or_shape.dtype,
                    "device": data_or_shape.device,
                    "layout": data_or_shape.layout,
                }
            )
        else:
            shape = tuple(data_or_shape)

        if device is not None:
            _init_kwargs["device"] = device

        if dtype is not None:
            _init_kwargs["dtype"] = dtype

        if delta_ndim is not None:
            offset = len(shape) - delta_ndim if delta_ndim >= 0 else abs(delta_ndim)
            shape = shape[offset:]

        self.delta = nn.Parameter(tr.zeros(shape, **_init_kwargs))  # type: ignore
        del _init_kwargs

        if init_fn is not None:
            init_fn(self.delta, **init_fn_kwargs)
        elif init_fn_kwargs:
            raise TypeError(
                f"No `init_fn` was specified, but the keyword arguments "
                f"{init_fn_kwargs} were provided."
            )

    def forward(self, data: Tensor) -> Tensor:
        """Add perturbation to data.

        Parameters
        ----------
        data: Tensor
            The data to perturb.

        Returns
        -------
        Tensor
            The perturbed data
        """
        return data + self.delta



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from .callbacks import MetricsCallback
from .launchers import HydraDDP

__all__ = ["MetricsCallback", "HydraDDP"]



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from pathlib import Path
from typing import Union

PathLike = Union[str, Path]

__all__ = ["md5_check"]


def md5_check(fname: PathLike, chunksize: int = 1024 ** 2) -> str:
    """Reads in data from disk and returns md5 hash"""
    import hashlib

    hash_md5 = hashlib.md5()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(chunksize), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from typing import Optional

import torch as tr
import torch.nn.functional as F

from rai_toolbox._typing import ArrayLike


def jensen_shannon_divergence(
    *probs: ArrayLike, weight: Optional[float] = None
) -> tr.Tensor:
    """
    Computes the Jensen-Shannon divergence [1]_ between n distributions:

    :math:`JSD(P_1, P_2, ..., P_n)`

    This loss is symmetric and is bounded by :math:`0 <= JSD(P_1, P_2, ..., P_n) <= \ln(n)`

    Parameters
    ----------
    probs : ArrayLike, shape-(N, D)
        A collection of n probability distributions. Each conveys of batch of N
        distributions over D categories.

    weight : Optional[float]
        A scaling factor that will be applied to the consistency loss.


    Returns
    -------
    loss : tr.Tensor, shape-(,)
        The scalar loss computed via the batch-mean.

    Notes
    -----
    The JSD loss is computed for each corresponding n-tuple of distributions and
    the batch-mean is ultimately returned.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Jensen%E2%80%93Shannon_divergence

    Examples
    --------
    Let's measure the divergence between three discrete distributions of length-two.

    >>> from rai_toolbox.losses import jensen_shannon_divergence
    >>> P1 = [[0.0, 1.0]]
    >>> P2 = [[1.0, 0.0]]
    >>> P3 = [[0.5, 0.5]]
    >>> jensen_shannon_divergence(P1, P2, P3)
    tensor(0.4621)

    The divergence is symmetric.

    >>> jensen_shannon_divergence(P1, P3, P2)
    tensor(0.4621)
    """

    probs = tuple(tr.as_tensor(p) for p in probs)

    if len(probs) < 2 or any(
        not isinstance(p, tr.Tensor) or p.dim() != 2 for p in probs
    ):
        raise ValueError(
            f"*probs must consist of at least two Tensors, and each tensor must have a shape of (N, D). Got {probs}"
        )

    zero = tr.tensor(0.0).type_as(probs[0])
    # Clamp mixture distribution to avoid exploding KL divergence
    log_p_mixture = tr.clamp(sum(probs, zero) / len(probs), 1e-7, 1).log()
    loss = sum(
        (F.kl_div(log_p_mixture, p, reduction="batchmean") for p in probs), zero
    ) / len(probs)

    if weight is not None:
        loss = loss * weight

    return loss



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from typing import Any, Dict, List, Optional, Union

import numpy as np
import torch
from torch import Generator, Tensor, default_generator
from torch.optim import SGD

from rai_toolbox._typing import Optimizer as Opt
from rai_toolbox._typing import OptimizerType, OptimParams, Partial
from rai_toolbox._utils import check_param_group_value, value_check

from .optimizer import REQUIRED, DatumParamGroup, ParamTransformingOptimizer

__all__ = ["ClampedGradientOptimizer", "TopQGradientOptimizer"]


class ClampedParamGroup(DatumParamGroup):
    clamp_min: Optional[float]
    clamp_max: Optional[float]


class _ClampedOptim(ParamTransformingOptimizer):
    param_groups: List[ClampedParamGroup]

    def __init__(
        self,
        params: Optional[OptimParams] = None,
        InnerOpt: Union[Opt, Partial[Opt], OptimizerType] = SGD,
        *,
        clamp_min: Optional[float] = None,
        clamp_max: Optional[float] = None,
        defaults: Optional[Dict[str, Any]] = None,
        param_ndim: Optional[int] = None,
        **inner_opt_kwargs,
    ) -> None:
        """
        Parameters
        ----------
        params : Sequence[Tensor] | Iterable[Mapping[str, Any]]
            Iterable of parameters or dicts defining parameter groups.

        InnerOpt : Type[Optimizer] | Partial[Optimizer], optional (default=`torch.nn.optim.SGD`)
            The optimizer that updates the parameters after their gradients have
            been transformed.

        epsilon : float
            Specifies the size of the L2-space ball that all parameters will be
            projected into, post optimization step.

        clamp_min: Optional[float]
            Lower-bound of the range to be clamped to.  Must be specified if `clamp_max` is `None`.

        clamp_max: Optional[float]
            Upper-bound of the range to be clamped to. Must be specified if `clamp_min`
            is `None`.

        grad_scale : float, optional (default=1.0)
            Multiplies each gradient in-place after the in-place transformation is
            performed. This can be specified per param-group.

        grad_bias : float, optional (default=0.0)
            Added to each gradient in-place after the in-place transformation is
            performed. This can be specified per param-group.

        defaults : Optional[Dict[str, Any]]
            Specifies default parameters for all parameter groups.

        param_ndim : Optional[int]
            Controles how `_pre_step_transform_` and `_post_step_transform_`  are
            broadcast onto a given parameter. This has no effect for
            `ClampedGradientOptimizer` and `ClampedParameterOptimizer`.

        **inner_opt_kwargs : Any
            Named arguments used to initialize `InnerOpt`.
        """
        if defaults is None:
            defaults = {}
        defaults.setdefault("clamp_min", clamp_min)
        defaults.setdefault("clamp_max", clamp_max)
        super().__init__(
            params,
            InnerOpt,
            defaults=defaults,
            param_ndim=param_ndim,
            **inner_opt_kwargs,
        )

        for group in self.param_groups:
            if group["clamp_min"] is None and group["clamp_max"] is None:
                raise ValueError("Either `clamp_min` or `clamp_max` must be specified")

            if group["clamp_min"] is not None and group["clamp_max"] is not None:
                value_check(
                    "clamp_min",
                    group["clamp_min"],
                    max_=group["clamp_max"],
                    upper_name="max_clamp",
                )


class ClampedGradientOptimizer(_ClampedOptim):
    """A gradient-tranforming  optimizer that clamps the elements of a gradient to
    fall within user-specified bounds **prior** to using `InnerOp.step` to update the
    corresponding parameter.

    See Also
    --------
    ClampedParameterOptimizer

    Examples
    --------
    Let's clamp each element of the parameter's gradient to `[-1, 3]` prior to
    performing a step with `SGD` using a learning rate of `1.0`.

    >>> import torch as tr
    >>> from rai_toolbox.optim import ClampedGradientOptimizer
    >>> x = tr.ones(2, requires_grad=True)
    >>> optim = ClampedGradientOptimizer(params=[x], lr=1.0, clamp_min=-1.0, clamp_max=3.0)

    >>> x.backward(gradient=tr.tensor([-0.5, 10]))
    >>> optim.step()
    >>> x.grad
    tensor([-0.5000,  3.0000])
    >>> x
    tensor([ 1.5000, -2.0000], requires_grad=True)"""

    def _pre_step_transform_(
        self, param: Tensor, optim_group: ClampedParamGroup
    ) -> None:
        if param.grad is None:  # pragma: no cover
            return
        param.grad.clamp_(min=optim_group["clamp_min"], max=optim_group["clamp_max"])


class ClampedParameterOptimizer(_ClampedOptim):
    """A parameter optimizer that clamps the elements of a parameter to fall within
    user-specified bounds **after** `InnerOp.step` has updated the parameter

    See Also
    --------
    ClampedGradientOptimizer

    Examples
    --------
    Let's perform a step with `SGD` using a learning rate of `1.0` to each of our parameters and then clamp their parameters to `[-1.0, 3.0]`.

    >>> import torch as tr
    >>> from rai_toolbox.optim import ClampedParameterOptimizer
    >>> x = tr.ones(2, requires_grad=True)
    >>> optim = ClampedParameterOptimizer(params=[x], lr=1.0, clamp_min=-1.0, clamp_max=3.0)

    >>> x.backward(gradient=tr.tensor([0.5, -10.0]))
    >>> optim.step()
    >>> x
    tensor([0.5000, 3.0000], requires_grad=True)"""

    def _post_step_transform_(
        self, param: Tensor, optim_group: ClampedParamGroup
    ) -> None:
        param.clamp_(min=optim_group["clamp_min"], max=optim_group["clamp_max"])


class TopQGradientOptimizer(ParamTransformingOptimizer):
    """A gradient-tranforming  optimizer that zeros the elements of a gradient whose
    absolue magnitudes fall below the Qth percentile. `InnerOp.step` is then to update
    the corresponding parameter.

    See Also
    --------
    L1qNormedGradientOptim
    ParamTransformingOptimizer"""

    def __init__(
        self,
        params: OptimParams,
        InnerOpt: Union[Partial[Opt], OptimizerType] = SGD,
        *,
        q: float = REQUIRED,
        dq: float = 0.0,
        param_ndim: Union[int, None] = -1,
        defaults: Optional[Dict[str, Any]] = None,
        generator: Generator = default_generator,
        **inner_opt_kwargs,
    ):

        r"""
        Parameters
        ----------
        params : Sequence[Tensor] | Iterable[Mapping[str, Any]]
            Iterable of parameters or dicts defining parameter groups.

        InnerOpt : Type[Optimizer] | Partial[Optimizer], optional (default=`torch.nn.optim.SGD`)
            The optimizer that updates the parameters after their gradients have
            been transformed.

        q : float
            Specifies the (fractional) percentile of absolute-largest gradient elements
            to retain when sparsifying the gradient. E.g., `q=0.9` means that only the
            gradient elements within the 90th-percentile will be retained.

            Must be within `[0.0, 1.0]`. The sparsification is applied to the gradient
            in accordance to `param_ndim`.

        dq : float, optional (default=0.0)
            If specified, the sparsity factor for each gradient transformation will
            be drawn from a uniform distribution over :math:`[q - dq, q + dq] \in [0.0, 1.0]`.

        param_ndim : Union[int, None], optional (default=-1)
            Determines how a parameter and its gradient is temporarily reshaped prior
            to being passed to both `_pre_step_transform_` and `_post_step_transform_`.
            By default,the transformation broadcasts over the tensor's first dimension
            in a batch-like style. This can be specified per param-group

            - A positive number determines the dimensionality of the tensor that the transformation will act on.
            - A negative number indicates the 'offset' from the dimensionality of the tensor (see "Notes" for examples).
            - `None` means that the transformation will be applied directly to the tensor without any broadcasting.

            See `ParamTransformingOptimizer` for more details and examples.

        grad_scale : float, optional (default=1.0)
            Multiplies each gradient in-place after the in-place transformation is
            performed. This can be specified per param-group.

        grad_bias : float, optional (default=0.0)
            Added to each gradient in-place after the in-place transformation is
            performed. This can be specified per param-group.

        defaults : Optional[Dict[str, Any]]
            Specifies default parameters for all parameter groups.

        generator : torch.Generator, optional (default=`torch.default_generator`)
            Controls the RNG source.

        **inner_opt_kwargs : Any
            Named arguments used to initialize `InnerOpt`.

        Examples
        --------
        Let's use `TopQGradientOptimizer` along with a standard SGD-step with a learning
        rate of `1.0`. We'll sparsify the gradient of a 2D parameter using varying
        percentile values. We set `param_ndim=None` so that no broadcasting occurs.

        >>> import torch as tr
        >>> from rai_toolbox.optim import TopQGradientOptimizer

        >>> gradient = tr.tensor([[0.5,   1.0],
        ...                       [-2.5, 0.30]])
        >>> for q in [0.0, 0.25, 0.5, 0.75, 1.0]:
        ...     x = tr.ones((2, 2), requires_grad=True)
        ...     optim = TopQGradientOptimizer(params=[x], lr=1.0, q=q, param_ndim=None)
        ...     x.backward(gradient=gradient)
        ...     optim.step()
        ...     print(f"grad (q={q})\n{x.grad}\nx:\n{x}\n---")
        grad (q=0.0)
        tensor([[ 0.5000,  1.0000],
                [-2.5000,  0.3000]])
        x:
        tensor([[0.5000, 0.0000],
                [3.5000, 0.7000]], requires_grad=True)
        ---
        grad (q=0.25)
        tensor([[ 0.5000,  1.0000],
                [-2.5000,  0.0000]])
        x:
        tensor([[0.5000, 0.0000],
                [3.5000, 1.0000]], requires_grad=True)
        ---
        grad (q=0.5)
        tensor([[ 0.0000,  1.0000],
                [-2.5000,  0.0000]])
        x:
        tensor([[1.0000, 0.0000],
                [3.5000, 1.0000]], requires_grad=True)
        ---
        grad (q=0.75)
        tensor([[ 0.0000,  0.0000],
                [-2.5000,  0.0000]])
        x:
        tensor([[1.0000, 1.0000],
                [3.5000, 1.0000]], requires_grad=True)
        ---
        grad (q=1.0)
        tensor([[0., 0.],
                [0., 0.]])
        x:
        tensor([[1., 1.],
                [1., 1.]], requires_grad=True)
        ---

        We'll repeat this exercise using `param_ndim=1` so that the top-Q
        sparsification is applied to each row independently (i.e. it is "broadcast"
        over each 1D sub-tensor in our gradient).

        >>> gradient = tr.tensor([[0.5,   1.0],
        ...                       [-2.5, 0.30]])
        >>> for q in [0.0, 0.5, 1.0]:
        ...     x = tr.ones((2, 2), requires_grad=True)
        ...     optim = TopQGradientOptimizer(params=[x], lr=1.0, q=q, param_ndim=1)
        ...     x.backward(gradient=gradient)
        ...     optim.step()
        ...     print(f"grad (q={q})\n{x.grad}\nx:\n{x}\n---")
        grad (q=0.0)
        tensor([[ 0.5000,  1.0000],
                [-2.5000,  0.3000]])
        x:
        tensor([[0.5000, 0.0000],
                [3.5000, 0.7000]], requires_grad=True)
        ---
        grad (q=0.5)
        tensor([[ 0.0000,  1.0000],
                [-2.5000,  0.0000]])
        x:
        tensor([[1.0000, 0.0000],
                [3.5000, 1.0000]], requires_grad=True)
        ---
        grad (q=1.0)
        tensor([[0., 0.],
                [0., 0.]])
        x:
        tensor([[1., 1.],
                [1., 1.]], requires_grad=True)
        ---
        """
        if defaults is None:
            defaults = {}
        defaults.setdefault("q", q)
        defaults.setdefault("dq", dq)

        super().__init__(
            params,
            InnerOpt=InnerOpt,
            defaults=defaults,
            param_ndim=param_ndim,
            **inner_opt_kwargs,
        )
        check_param_group_value(
            "q", self.param_groups, optional=True, min_=0.0, max_=1.0
        )
        check_param_group_value(
            "dq", self.param_groups, optional=True, min_=0.0, max_=1.0
        )
        self._generator = value_check("generator", generator, type_=torch.Generator)

    def _pre_step_transform_(self, param: Tensor, optim_group: Dict[str, Any]) -> None:

        if param.grad is None:  # pragma: no cover
            return

        q = optim_group["q"]
        dq = optim_group["dq"]

        if dq > 0.0:
            _qlow = max(0.0, q - dq)
            _qhigh = min(1.0, q + dq)

            q = float(
                (_qhigh - _qlow) * torch.rand(1, generator=self._generator) + _qlow
                if dq and (_qlow < q or _qhigh > q)
                else q
            )

        # Convert percent to number of entries
        shp = param.grad.shape

        num_q = 1.0 - q
        num_q = round(num_q * np.prod(shp[1:]))

        g = param.grad.flatten(1)
        s = torch.zeros_like(g)

        if num_q:
            _, corners_q = torch.topk(g.abs(), num_q, dim=1)
            batch_idx = torch.tensor([[i] * num_q for i in range(shp[0])])
            s[batch_idx, corners_q] = g[batch_idx, corners_q]

        param.grad[...] = s.view(shp)  # type: ignore



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from . import _version
from ._utils import to_batch
from ._utils.stateful import evaluating, freeze, frozen
from .losses._utils import negate

__all__ = ["evaluating", "frozen", "freeze", "negate", "to_batch"]

__version__ = _version.get_versions()["version"]

del _version



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from ._utils import load_experiment, load_from_checkpoint
from .lightning import HydraDDP, MetricsCallback
from .workflows import (
    BaseWorkflow,
    MultiRunMetricsWorkflow,
    RobustnessCurve,
    hydra_list,
    multirun,
)

__all__ = [
    "load_experiment",
    "load_from_checkpoint",
    "MetricsCallback",
    "MultiRunMetricsWorkflow",
    "HydraDDP",
    "RobustnessCurve",
    "BaseWorkflow",
    "multirun",
    "hydra_list",
]



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from typing import Optional, Union

import torch
from torch import Generator, Tensor, default_generator

from rai_toolbox._utils import to_batch as _to_batch
from rai_toolbox._utils import validate_param_ndim as _validate_param_ndim
from rai_toolbox._utils import value_check


@torch.no_grad()
def uniform_like_l1_n_ball_(
    x: Tensor,
    epsilon: float = 1,
    param_ndim: Union[int, None] = -1,
    generator: Generator = default_generator,
) -> None:
    r"""Uniform sampling of an :math:`\epsilon`-sized `n`-ball for :math:`L^1`-norm, where `n` is controlled by `x.shape` and `param_ndim`. The result overwrites `x` in-place.

    Parameters
    ----------
    x: Tensor, shape-(N, D, ...)
        The tensor from which to generate a new random tensor, i.e., returns a tensor of
        similar shape and on the same device.

        By default, each of the `N` shape-`(D, ...)` subtensors are initialized
        independently. See `param_ndim` to contol this behavior.

    epsilon : float, optional (default=1)
        Determines the radius of the ball.

    param_ndim : int | None, optional (default=-1)
        Determines the dimenionality of the subtensors that are sampled by this
        function.

        - A positive number determines the dimensionality of each subtensor to be drawn and packed into the shape-`(N, D, ...)` resulting tensor.
        - A negative number determines from the dimensionality of the subtensor in terms of the offset of `x.ndim`. E.g. -1 indicates that `x` is arranged in a batch-style, and that `N` independent shape-`(D, ...)` tensors will be sampled.
        - `None` means that a single tensor of shape-`(N, D, ...)` is sampled.

    Returns
    -------
    out: Tensor, shape-(N, D, ...)
        A new random tensor of the same shape and on the same device as the input.

    References
    ----------
    .. [1] Rauber et al., 2020, Foolbox Native: Fast adversarial attacks to benchmark the robustness of machine learning models in PyTorch, TensorFlow, and JAX https://doi.org/10.21105/joss.02607

    .. [2] https://mathoverflow.net/a/9188

    Examples
    --------
    >>> import torch as tr
    >>> from rai_toolbox.perturbations.init import uniform_like_l1_n_ball_

    Drawing two shape-`(3,)` tensors from an :math:`\epsilon=2` sized :math:`L^1` 3D-ball.

    >>> x = tr.zeros(2, 3)
    >>> uniform_like_l1_n_ball_(x, epsilon=2.0) # default: param_ndim=-1
    >>> x
    tensor([[0.3301, 0.8459, 0.7071],
            [0.3470, 0.5077, 0.0977]])
    >>> tr.linalg.norm(x, dim=1, ord=1) < 2.0
    tensor([True, True])

    Drawing one tensor shape-`(6,)` tensor from a :math:`\epsilon=2` sized :math:`L^1` 6D-ball, and
    storing it in `x` as a shape-`(2, 3)` tensor. We specify `param_ndim=2` (or
    `param_ndim=None`) to achieve this.

    >>> x = tr.zeros(2, 3)
    >>> uniform_like_l1_n_ball_(x, epsilon=2.0, param_ndim=2)
    >>> x
    tensor([[0.1413, 0.5270, 0.1570],
            [0.4817, 0.2760, 0.4076]])
    >>> tr.linalg.norm(x, ord=1) < 2.0
    tensor(True)
    """
    _validate_param_ndim(param_ndim=param_ndim, p=x)
    value_check("epsilon", epsilon, min_=0.0, incl_min=False)

    xflat = _to_batch(x, param_ndim=param_ndim).flatten(1)
    nbatch, ndim = xflat.shape
    u = xflat.new(nbatch, ndim).uniform_(generator=generator)
    v = u.sort(dim=1).values
    vp = torch.cat((xflat.new_zeros(nbatch, 1), v[:, : ndim - 1]), dim=1)
    assert v.shape == vp.shape
    z = v - vp
    sign = xflat.new(nbatch, ndim).uniform_().sign()
    x.copy_(epsilon * (sign * z).reshape(x.shape))


@torch.no_grad()
def uniform_like_l2_n_ball_(
    x: Tensor,
    epsilon: float = 1,
    param_ndim: Union[int, None] = -1,
    generator: Generator = default_generator,
) -> None:
    r"""Uniform sampling within an :math:`\epsilon`-sized `n`-ball for :math:`L^2`-norm, where `n`
    is controlled by `x.shape` and `param_ndim`. The result overwrites `x` in-place.

    Parameters
    ----------
    x: Tensor, shape-(N, D, ...)
        The tensor to generate a new random tensor from, i.e., returns a tensor of
        similar shape and on the same device.

        By default, each of the `N` shape-`(D, ...)` subtensors are initialized
        independently. See `param_ndim` to contol this behavior.

    epsilon : float, optional (default=1)
        Determines the radius of the ball.

    param_ndim : int | None, optional (default=-1)
        Determines the dimenionality of the subtensors that are sampled by this
        function.

        - A positive number determines the dimensionality of each subtensor to be drawn and packed into the shape-`(N, D, ...)` resulting tensor.
        - A negative number determines from the dimensionality of the subtensor in terms of the offset of `x.ndim`. E.g. -1 indicates that `x` is arranged in a batch-style, and that `N` independent shape-`(D, ...)` tensors will be sampled.
        - `None` means that a single tensor of shape-`(N, D, ...)` is sampled.

    generator : torch.Generator, optional (default=`torch.default_generator`)
        Controls the RNG source.

    Returns
    -------
    out: Tensor, shape-`(N, ...)`
        A new random tensor of the same shape and on the same device as the input.

    References
    ----------
    .. [1] Rauber et al., 2020, Foolbox Native: Fast adversarial attacks to benchmark the robustness of machine learning models in PyTorch, TensorFlow, and JAX https://doi.org/10.21105/joss.02607

    .. [2] Voelker et al., 2017, Efficiently sampling vectors and coordinates from the n-sphere and n-ball http://compneuro.uwaterloo.ca/files/publications/voelker.2017.pdf

    .. [3] Roberts, Martin, 2020, How to generate uniformly random points on n-spheres and in n-balls http://extremelearning.com.au/how-to-generate-uniformly-random-points-on-n-spheres-and-n-balls/

    Examples
    --------
    >>> import torch as tr
    >>> from rai_toolbox.perturbations.init import uniform_like_l2_n_ball_

    Drawing two shape-`(3,)` tensors from an :math:`\epsilon=2`-sized :math:`L^2` 3D-ball.

    >>> x = tr.zeros(2, 3)
    >>> uniform_like_l2_n_ball_(x, epsilon=2.0, param_ndim=1)
    >>> x
    tensor([[0.3030, 1.4269, 0.3927],
            [1.4015, 0.4913, 1.3028]])
    >>> tr.linalg.norm(x, dim=1, ord=2) < 2.0
    tensor([True, True])

    Drawing one shape-`(6, )` tensor from a :math:`\epsilon=2`-sized :math:`L^2` 6D-ball, and
    storing it in `x` as a shape-`(2, 3)` tensor. We specify `param_ndim=2` (or
    `param_ndim=None`) to achieve this.

    >>> x = tr.zeros(2, 3)
    >>> uniform_like_l2_n_ball_(x, epsilon=2.0, param_ndim=2)
    >>> x
    tensor([[-0.6903, -0.8597,  0.0109],
            [ 0.0906, -0.2387, -0.3059]])
    >>> tr.linalg.norm(x, ord=2) < 2.0
    tensor(True)
    """
    _validate_param_ndim(param_ndim=param_ndim, p=x)
    value_check("epsilon", epsilon, min_=0.0, incl_min=False)

    xflat = _to_batch(x, param_ndim=param_ndim).flatten(1)
    nbatch, ndim = xflat.shape
    z = xflat.new(nbatch, ndim + 2).normal_(generator=generator)
    r = z.norm(p=2, dim=1, keepdim=True)  # type: ignore
    x.copy_(epsilon * (z / r)[:, :ndim].reshape(x.shape))


@torch.no_grad()
def uniform_like_linf_n_ball_(
    x: Tensor,
    epsilon: float = 1,
    param_ndim: Optional[int] = None,
    generator: Generator = default_generator,
) -> None:
    r"""Uniform sampling within an :math:`\epsilon`-sized `n`-ball for :math:`L^{\infty}`-norm.
    The result overwrites `x` in-place.

    Parameters
    ----------
    x: Tensor, shape-(N, D, ...)
        The tensor to generate a new random tensor from, i.e., returns a tensor of
        similar shape and on the same device.

    epsilon : float, optional (default=1)
        Determines the radius of the ball.

    param_ndim : Optional[int]
        Unused. Included for parity with other init functions.

    generator : torch.Generator, optional (default=`torch.default_generator`)
        Controls the RNG source.

    Returns
    -------
    out: Tensor, shape-(N, ...)
        A new random tensor of the same shape and on the same device as the input.

    Examples
    --------
    >>> import torch as tr
    >>> from rai_toolbox.perturbations.init import uniform_like_linf_n_ball_
    >>> x = tr.zeros(2, 3)
    >>> uniform_like_linf_n_ball_(x, epsilon=2.0)
    >>> x
    tensor([[ 1.7092, -1.8723, -0.0806],
            [-1.4680, -1.8782, -0.1998]])
    >>> x.abs() < 2.
    tensor([[True, True, True],
            [True, True, True]])
    """
    del param_ndim
    value_check("epsilon", epsilon, min_=0.0, incl_min=False)

    x.uniform_(-epsilon, epsilon, generator=generator)



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import inspect
from abc import ABCMeta
from typing import Any, Callable, Dict, List, Optional, TypeVar, Union, cast, overload

import torch
from torch import Tensor
from torch.optim import SGD, Optimizer
from typing_extensions import TypedDict

from rai_toolbox._typing import InstantiatesTo
from rai_toolbox._typing import Optimizer as Opt
from rai_toolbox._typing import OptimizerType, OptimParams, Partial, instantiates_to
from rai_toolbox._utils import validate_param_ndim as _validate_param_ndim

_T = TypeVar("_T", bound=Optional[Union[Tensor, float]])

__all__ = ["ParamTransformingOptimizer", "ChainedParamTransformingOptimizer"]


REQUIRED: Any = inspect.signature(SGD).parameters["lr"].default


class DatumParamGroup(TypedDict):
    params: List[Tensor]
    param_ndim: Optional[int]
    grad_scale: float
    grad_bias: float


def _shares_memory(x: Tensor, y: Tensor) -> bool:
    return x.storage().data_ptr() == y.storage().data_ptr()


def _reshape_to_batch(x: Tensor, param_ndim: Optional[int]) -> Tensor:
    """Reshapes to a shape-`(N, d1, ..., dm)`, where `(d1, ..., dm)` has `param_ndim`
    dimensions. Dimensions will be added or consolidated to achieve this."""
    if param_ndim is None:
        param_ndim = x.ndim

    if param_ndim < 0:
        param_ndim += x.ndim

    # `1 + param_ndim` is the required dimensionality
    # for a shape-(N, d0, d1, ...) tensor, where (d0, d1, ...)
    # is the shape of the param_ndim-dimension tensor.
    #
    # We compute `ndim_delta` to determine if we need to add
    # or consolidate dimensions to create the shape-(N, d0, d1, ...)
    # tensor.
    ndim_delta = (1 + param_ndim) - x.ndim

    if ndim_delta > 0:
        # E.g.:
        #   p.shape: (d0, )
        #   desired shape: (N=1, d0)
        x = x[ndim_delta * (None,)]
    elif ndim_delta < 0:
        # E.g.:
        #   p.shape: (d0, d1, d2, d3)
        #   desired shape: (N=d0*d1, d2, d3)
        x = x.view(-1, *x.shape[x.ndim - param_ndim :])
    if x.ndim < 2:  # make at least 2D
        # (N,) -> (N, 1)
        x = x.view(*x.shape, *(1,) * (2 - x.ndim))
    return x


def _to_batch(p: Tensor, param_ndim: Optional[int]) -> Tensor:
    """
    Returns a view of `p`, reshaped as shape-(N, d0, ...) where (d0, ...)
    has `param_ndim` entries.

    See Parameters for further description

    Parameters
    ----------
    p : Tensor

    param_ndim: Optional[int]
        Determines the shape of the resulting parameter

        - A positive number determines the dimensionality of the tensor that the transformation will act on.
        - A negative number indicates the 'offset' from the dimensionality of the tensor.
        - `None` means that the transformation will be applied to the tensor without any broadcasting.

    Returns
    -------
    reshaped_p: Tensor

    Examples
    --------
    >>> import torch as tr
    >>> x = tr.rand((3, 5, 2))

    >>> _to_batch(x, 0).shape
    torch.Size([30, 1])

    >>> _to_batch(x, 1).shape
    torch.Size([15, 2])

    >>> _to_batch(x, 2).shape
    torch.Size([3, 5, 2])

    >>> _to_batch(x, None).shape
    torch.Size([1, 3, 5, 2])

    >>> _to_batch(x, -1).shape
    torch.Size([3, 5, 2])

    >>> _to_batch(x, -2).shape
    torch.Size([15, 2])

    >>> _to_batch(x, -3).shape
    torch.Size([30, 1])
    """

    # atleast_2d needed for case where p was scalar
    vp = _reshape_to_batch(p, param_ndim=param_ndim)

    if p.grad is not None:
        vp.grad = _reshape_to_batch(p.grad, param_ndim=param_ndim)

    # vp (vp.grad) must be a view of p (p.grad). There is
    # not a simple way to assert this.

    # our views must be size-preserving
    assert torch.numel(vp) == torch.numel(p)
    return vp


class ParamTransformingOptimizer(Optimizer, metaclass=ABCMeta):
    r"""An optimizer that performs an in-place transformation to
    each parameter, both before and after performing the gradient-based update on each
    parameter via `InnerOptim.step`::

       _pre_step_transform_(param)
       param = InnerOptim.step(param, ...)
       _post_step_transform_(param)

    Note that `_pre_step_transform_` and `_post_step_transform_` can be used to update
    a parameter and/or its gradient. Also, this optimizer exposes `param_ndim` as a
    means of controlling how these transforms broadcast (if at all) over any given
    tensor.

    Notes
    -----
    `ParamTransformingOptimizer` mirrors state with `InnerOpt` so that their
    `param_groups`, `defaults`, and `state` are always in sync.

    `ParamTransformingOptimizer` is designed to be combined with other,
    standard gradient-based optimizers (e.g., Adam) via encapsulation, rather than
    through inheritance. I.e., `ParamTransformingOptimizer(InnerOpt=<...>)` will apply
    `_pre_step_transform_` on a parameter, and then use `InnerOpt.step(...)` to update
    said parameter, and finally will apply `_post_step_transform_` to the parameter.

    If a closure is supplied to the `.step(...)` method, then the `_pre_step_transform_`
    is applied after the closure call and prior to the parameter steps.

    Methods
    -------
    _pre_step_transform_
    _post_step_transform_
    project

    See Also
    --------
    ChainedParamTransformingOptimizer
    """

    param_groups: List[DatumParamGroup]

    def __init__(
        self,
        params: Optional[OptimParams] = None,
        InnerOpt: Union[Opt, Partial[Opt], OptimizerType] = SGD,
        *,
        param_ndim: Union[int, None] = -1,
        grad_scale: float = 1.0,
        grad_bias: float = 0.0,
        defaults: Optional[Dict[str, Any]] = None,
        **inner_opt_kwargs,
    ) -> None:
        r"""
        Parameters
        ----------
        params : Sequence[Tensor] | Iterable[ParamGroup]
            Iterable of parameters to optimize or dicts defining parameter groups

        InnerOpt : Type[Optimizer] | Partial[Optimizer], optional (default=`torch.nn.optim.SGD`)
            The optimizer that updates the parameters after their gradients have
            been transformed.

        param_ndim : int | None, optional (default=-1)
            Determines how a parameter and its gradient are temporarily reshaped prior
            to being passed to both `_pre_step_transform_` and `_post_step_transform_`.
            By default, the transformation broadcasts over the tensor's first dimension
            in a batch-like style.

            - A positive number determines the dimensionality of the tensor that the transformation will act on.
            - A negative number indicates the 'offset' from the dimensionality of the tensor (see "Notes" for examples).
            - `None` means that the transformation will be applied directly to the tensor without any broadcasting.

            See "Notes" for more details.

        grad_scale : float, optional (default=1.0)
            Multiplies each gradient in-place after the pre-step transformation is
            performed. This can be specified per param-group.

        grad_bias : float, optional (default=0.0)
            Added to each gradient in-place after the pre-step transformation is
            performed. This can be specified per param-group.

        defaults : Optional[Dict[str, Any]]
            Specifies default parameters for all parameter groups.

        **inner_opt_kwargs : Any
            Named arguments used to initialize `InnerOpt`.

        Notes
        -----
        .. _param-ndim-add:

        **Additional Explanation of `param_ndim`**

        Consider a parameter of shape `(d0, d1, d2, d4)`.

        If `param_ndim=0`, then the parameter and its gradient will be temporarily
        reshaped to a shape-`(d0 * d1 * d2 * d3, 1)` so that the transformation will be
        applied elementwise to the tensor.

        If `param_ndim=1` (or `param_ndim=-3`), then the parameter and its gradient
        will be temporarily reshaped to a shape-`(d0 * d1 * d2, d3)` so that the
        transformation will be broadcast over each shape-`(d3,)` sub-tensor.

        If `param_ndim=2` (or `param_ndim=-2`), then the parameter and its gradient
        will be temporarily reshaped to a shape-`(d0 * d1, d2, d3)` so that the
        transformation will be broadcast over each shape-`(d2, d3)` sub-tensor.

        If `param_ndim=3` (or `param_ndim=-1`), then the parameter and its gradient
        will be temporarily reshaped to a shape-`(d0, d1, d2, d3)` so that the
        transformation will be broadcast over each shape-`(d1, d2, d3)` sub-tensor.

        If `param_ndim=4` (or `param_ndim=None`), then the parameter and its gradient
        will be temporarily reshaped to a shape-`(1, d0, d1, d2, d3)` so that the
        transformation will be applied to the shape-`(d0, d1, d2, d3)` tensor without
        broadcasting.

        Examples
        --------
        **Creating a gradient-transforming optimizer**

        Let's create a gradient-transforming optimizer that replaces the gradient
        of each parameter with the elementwise sign of the gradient (:math:`\pm 1`)
        prior to performing the step of the inner optimizer:

        >>> import torch as tr
        >>> from rai_toolbox.optim import ParamTransformingOptimizer
        >>> class SignedGradientOptim(ParamTransformingOptimizer):
        ...
        ...     def _pre_step_transform_(self, param: tr.Tensor, **_kwds) -> None:
        ...         if param.grad is None:
        ...             return
        ...         tr.sign(param.grad, out=param.grad)  # operates in-place

        Now we'll use this optimizer – with `torch.optim.AdamW` providing the actual
        parameter-update functionality – to update the parameter.

        >>> x = tr.tensor([-10.0, 10.0], requires_grad=True)
        >>> optim = SignedGradientOptim([x], InnerOpt=tr.optim.AdamW, lr=0.1)

        Using `x` in a calculation and compute an associated gradient for it:

        >>> (10_000 * x).sum().backward()

        Updating `x` using our grad-sign + AdamW optimizer:

        >>> optim.step()
        >>> x
        tensor([-10.9000,   8.9000], requires_grad=True)

        This was a simple optimizer which did not involve any broadcasting in the
        gradient transformation; the next example will involve broadcasting.

        **Controlling the gradient transformation with param_ndim**

        To understand the role of `param_ndim` let's design an optimizer that
        normalizes a parameter's gradient by its max value – along some user-specified dimension – prior to performing the gradient-based update to its parameter.

        >>> class MaxNormedGradientOptim(ParamTransformingOptimizer):
        ...
        ...     def _pre_step_transform_(self, param: tr.Tensor, **_kwds) -> None:
        ...         if param.grad is None:
        ...             return
        ...
        ...         g = param.grad.flatten(1) # (N, d1, ..., dm) -> (N, d1 * ... * dm)
        ...         max_norms = tr.max(g, dim=1).values
        ...         max_norms = max_norms.view(-1, *([1] * (param.ndim - 1)))  # reshape to have dimenionality-m
        ...         param.grad /= tr.clamp(max_norms, 1e-20, None)  # clamp to prevent div by 0

        Note that we design `_pre_step_transform_` to operate in-place on the gradient
        and that we treat the gradient as if it has a shape `(N, d1, ..., dm)`, where we
        want to compute the max over each of the `N` sub-tensors of
        shape-`(d1, ..., dm)`.

        Critically, we did not use `param_ndim` at all in this method;
        `ParamTransformingOptimizer` assumes that we designed this method to broadcast
        in a batch-style, as we did, and it automatically leverages `param_ndim`
        to reshape the parameter and its gradient appropriately prior to calling
        `_pre_step_transform_`.

        Now we will create a shape-`(2, 2)` parameter to see how `MaxNormedGradientOptim`
        can compute the max-norm over various dimensions of the parameter. Let's print
        out the transformed gradient when we use each of `param_ndim`: `0`, `1`, or `2`.

        >>> x = tr.tensor([[1.0, 2.0],
        ...                [20.0, 10.0]], requires_grad=True)
        >>> for param_ndim in [0, 1, 2]:
        ...     optim = MaxNormedGradientOptim([x], param_ndim=param_ndim, InnerOpt=tr.optim.SGD, lr=0.0)
        ...
        ...     loss = (x * x).sum()
        ...     loss.backward()
        ...     optim.step()
        ...     print(f"param_ndim: {param_ndim}, normed grad:\n{x.grad}\n..")
        ...     optim.zero_grad()
        param_ndim: 0, normed grad:
        tensor([[1., 1.],
                [1., 1.]])
        ..
        param_ndim: 1, normed grad:
        tensor([[0.5000, 1.0000],
                [1.0000, 0.5000]])
        ..
        param_ndim: 2, normed grad:
        tensor([[0.0500, 0.1000],
                [1.0000, 0.5000]])

        See that `param_ndim=0` applies the max-norm elementwise, whereas `param_ndim=1`
        applied the max-norm to each 1D row of the gradient, and `param_ndim=2` applies
        the max-norm over the entire 2D gradient.

        **Creating a parameter-constraining optimizer**

        Let's create an optimizer that clamps each parameter's values so that
        they all fall within `[-1, 1]` after performing it's gradient-based step on the parameter.

        >>> import torch as tr
        >>> from rai_toolbox.optim import ParamTransformingOptimizer
        >>> class ClampedParamOptim(ParamTransformingOptimizer):
        ...     def _post_step_transform_(self, param: tr.Tensor, optim_group: dict) -> None:
        ...         param.clamp_(min=-1.0, max=1.0)  # note: clamp occurs in-place

        >>> x = tr.tensor([-10., 1.], requires_grad=True)
        >>> optim = ClampedParamOptim([x], lr=0.1)  # InnerOpt=SGD by default
        >>> x.backward(gradient=tr.tensor([-1., 1.]))
        >>> optim.step()  # parameters updated via SGD.step() and then clamped
        >>> x
        tensor([-1.0000,  0.9000], requires_grad=True)

        Note that this is a particularly simple function, which acts elementwise on
        each parameter, and thus does not require us to include `param_ndim` in the
        optimizer's param-groups.
        """
        if defaults is None:
            defaults = {}
        defaults.setdefault("param_ndim", param_ndim)
        defaults.setdefault("grad_scale", grad_scale)
        defaults.setdefault("grad_bias", grad_bias)

        if instantiates_to(InnerOpt, Optimizer):
            if params is None:
                raise TypeError(
                    "`params` cannot be `None` when `InnerOpt` is an un-instantiated "
                    "optimizer type."
                )
            super().__init__(params, defaults)  # type: ignore
            self.inner_opt = InnerOpt(self.param_groups, **inner_opt_kwargs)  # type: ignore
        elif isinstance(InnerOpt, Optimizer):
            self.inner_opt = InnerOpt
            super().__init__(self.inner_opt.param_groups, defaults)
        else:
            raise TypeError(
                f"`InnerOpt` must be an Optimizer type or instance, got: {InnerOpt}"
            )

        # ensure inner-opt's defaults include those of `self`
        self.inner_opt.defaults.update(
            **{
                k: v
                for k, v in self.inner_opt.defaults.items()
                if k not in self.defaults
            },
            **self.defaults,
        )

        # state of `self` must mirror that of inner-opt
        self.__setstate__(self.inner_opt.__getstate__())  # type: ignore

        for group in self.param_groups:
            param_ndim = group["param_ndim"]
            if param_ndim is not None and not isinstance(param_ndim, int):
                raise TypeError(
                    f"`param_ndim` must be an int or None, got: {param_ndim}"
                )

            if not isinstance(group["grad_scale"], (float, int)):
                raise TypeError(
                    f"grad_scale must be a float, got {group['grad_scale']}"
                )

            if not isinstance(group["grad_bias"], (float, int)):
                raise TypeError(f"grad_bias must be a float, got {group['grad_bias']}")

            for p in group["params"]:
                p: Tensor
                _validate_param_ndim(param_ndim=param_ndim, p=p)

    def state_dict(self) -> dict:
        return self.inner_opt.state_dict()

    def __setstate__(self, state: dict):
        self.inner_opt.__setstate__(state)
        super().__setstate__(self.inner_opt.__getstate__())  # type: ignore

    def __getstate__(self) -> dict:
        return self.inner_opt.__getstate__()  # type: ignore

    def __repr__(self) -> str:
        return super().__repr__().replace("(", f"[{type(self.inner_opt).__name__}](", 1)

    def _pre_step_transform_(
        self, param: Tensor, optim_group: DatumParamGroup
    ) -> None:  # pragma: no cover
        """Applies an in-place transform on each parameter in the given param group **before** that parameter has been updated via `InnerOpt.step`.

        This defaults to a no-op.

        Parameters
        ----------
        param : torch.Tensor, shape-(N, d0, ...)
            The parameter to be modified in-place.

            `param` and `param.grad` will have been reshaped to have a
            shape-`(N, d0, ...)` where `(d0, ...)` contains `param_ndim` entries.

        optim_group : Dict[str, Any]
            The parameter group associated with `param`.

        Notes
        -----
        This transform should *always* be designed to broadcast over the leading
        dimension of the tensor being modified. That is, each parameter/gradient should be assumed to have the shape-`(N, d0, ...)` and the transformation should be
        applied - in-place - to each shape-`(d0, ...)` sub-tensor.

        Prior to calling `_pre_step_transform_`, `ParamTransformingOptimizer`
        will temporarily reshape each parameter and its gradient to have the appropriate
        shape – in accordance with the value specified for `param_ndim` – such that
        the shape-`(d0, ...)` tensor contains `param_ndim` entries.

        In the case where `param_ndim=0`, the transformation will be applied to a
        shape-`(T, 1)` tensor, where `T` corresponds to the total number of elements
        in the tensor."""
        del param
        del optim_group
        return None

    def _post_step_transform_(
        self, param: Tensor, optim_group: DatumParamGroup
    ) -> None:  # pragma: no cover
        """Applies an in-place transform on each parameter in the given param group **after** that parameter has been updated via `InnerOpt.step`.

        This defaults to a no-op.

        Parameters
        ----------
        param : torch.Tensor, shape-(N, d0, ...)
            The parameter to be modified in-place.

            `param` and `param.grad` will have been reshaped to have a
            shape-`(N, d0, ...)` where `(d0, ...)` contains `param_ndim` entries.

        optim_group : Dict[str, Any]
            The parameter group associated with `param`.

        Notes
        -----
        This transform should *always* be designed to broadcast over the leading
        dimension of the tensor being modified. That is, each parameter/gradient should be assumed to have the shape-(N, d0, ...) and the transformation should be
        applied - in-place - to each shape-`(d0, ...)` sub-tensor.

        Prior to calling `_post_step_transform_`, `ParamTransformingOptimizer`
        will temporarily reshape each parameter and its gradient to have the appropriate
        shape – in accordance with the value specified for `param_ndim` – such that
        the shape-`(d0, ...)` tensor contains `param_ndim` entries.

        In the case where `param_ndim=0`, the transformation will be applied to a
        shape-`(T, 1)` tensor, where `T` corresponds to the total number of elements
        in the tensor.
        """
        del param
        del optim_group
        return None

    @torch.no_grad()
    def project(self) -> None:
        """Update each parameter in-place by calling `_post_step_transform_` on the
        parameter.

        `.project` is called automatically by `.step`."""
        for group in self.param_groups:
            param_ndim = group["param_ndim"]

            for p in group["params"]:
                p = _to_batch(p, param_ndim)
                self._post_step_transform_(param=p, optim_group=group)

    def _apply_pre_step_transform_(self):
        for group in self.param_groups:
            for p in group["params"]:
                p: Tensor
                orig_p = p
                if p.grad is None:
                    continue
                assert orig_p.grad is not None

                p = _to_batch(p, group["param_ndim"])
                assert p.grad is not None

                self._pre_step_transform_(p, optim_group=group)

                if group["grad_scale"] != 1.0:
                    p.grad *= group["grad_scale"]

                if group["grad_bias"] != 0.0:
                    p.grad += group["grad_bias"]

                if p.grad is None or not _shares_memory(orig_p.grad, p.grad):
                    raise ValueError(
                        f"`{type(self).__name__}._pre_step_transform_` did "
                        " not modify the gradient of the parameter in-place."
                        " \nNote that setting `p.grad` directly replaces the"
                        " tensor, rather than writing to the tensor."
                    )

    @torch.no_grad()
    def _create_closure(self, closure: Callable[[], _T]) -> Callable[[], Optional[_T]]:
        def new_closure():
            with torch.enable_grad():
                loss = closure()

            self._apply_pre_step_transform_()
            return loss

        return new_closure

    @overload
    def step(self, closure: Callable[[], _T]) -> _T:  # pragma: no cover
        ...

    @overload
    def step(self) -> None:  # pragma: no cover
        ...

    @overload
    def step(
        self, closure: Optional[Callable[[], _T]] = None
    ) -> Optional[_T]:  # pragma: no cover
        ...

    @torch.no_grad()
    def step(self, closure=None):
        if closure is not None:
            closure = self._create_closure(closure)
            loss = self.inner_opt.step(closure)  # type: ignore
        else:
            self._apply_pre_step_transform_()
            self.inner_opt.step()
            loss = None
        self.project()
        loss = cast(Optional[Union[float, Tensor]], loss)
        return loss


class ChainedParamTransformingOptimizer(ParamTransformingOptimizer):
    """Chains together an arbitrary number of parameter-transforming optimizers,
    composing their pre- and post-step transformation functions to modify the parameters (and their gradients) in-place. `InnerOpt.step()` applies the gradient-based update
    to each parameter.

    I.e., passing `Opt1, Opt2, ..., OptN` to `ChainedParamTransformingOptimizer` will
    update a parameter using: `OptN.fn_(...(Opt2.fn_(Opt1.fn_(param)))`,
    where `fn_` is a shorthand for `_pre_step_transform_` / `_post_step_transform_`.

    Notes
    -----
    `ChainedParamTransformingOptimizer` mirrors state with `InnerOpt`, and with all of
    the user-specified chained gradient-trasnformers, so that their `param_groups`,
    `defaults`, and `state` are always in sync.

    See Also
    --------
    ParamTransformingOptimizer
    """

    def __init__(
        self,
        *transforming_optimizers: InstantiatesTo[ParamTransformingOptimizer],
        params: Optional[OptimParams] = None,
        InnerOpt: Union[Opt, Partial[Opt], OptimizerType] = SGD,
        param_ndim: Union[int, None] = -1,
        grad_scale: float = 1,
        grad_bias: float = 0,
        defaults: Optional[Dict[str, Any]] = None,
        **inner_opt_kwargs,
    ) -> None:
        r"""
        Parameters
        ----------
        *transforming_optimizers: InstantiatesTo[ParamTransformingOptimizer],
            An arbitrary number of parameter-transforming optimizers, whose
            `_pre_step_transform_` and `_post_step_transform_` methods, respectively,
            will be composed from left to right –
            `Opt1, Opt2, ..., OptN -> fN_(...f2_(f1_(grad)))` – to modify a parameter prior to / after being updated by `InnerOpt.step`

        params : Optional[Sequence[Tensor] | Iterable[ParamGroup]]
            Iterable of parameters to optimize or dicts defining parameter groups

        InnerOpt : Type[Optimizer] | Partial[Optimizer], optional (default=`torch.nn.optim.SGD`)
            The optimizer that updates the parameters after `_pre_step_transform_` has been applied to each of them.

        param_ndim : int | None, optional (default=-1)
            Determines how a parameter and its gradient is temporarily reshaped prior
            to being passed to both `_pre_step_transform_` and `_post_step_transform_`.
            By default,the transformation broadcasts over the tensor's first dimension
            in a batch-like style.

            - A positive number determines the dimensionality of the tensor that the transformation will act on.
            - A negative number indicates the 'offset' from the dimensionality of the tensor (see "Notes" for examples).
            - `None` means that the transformation will be applied directly to the tensor without any broadcasting.

            See `ParamTransformingOptimizer` for more details and examples.

        grad_scale : float, optional (default=1.0)
            Multiplies each gradient in-place after the in-place transformation is
            performed. This can be specified per param-group.

        grad_bias : float, optional (default=0.0)
            Added to each gradient in-place after the in-place transformation is
            performed. This can be specified per param-group.

        defaults : Optional[Dict[str, Any]]
            Specifies default parameters for all parameter groups.

        **inner_opt_kwargs : Any
            Named arguments used to initialize `InnerOpt`.

        Examples
        --------
        **Basic Example**

        Let's chain together two gradient-transforming optimizers supplied by rAI-toolbox:
        `TopQGradientOptimizer` and `ClampedGradientOptimizer`

        >>> from rai_toolbox.optim import (
        ... ChainedParamTransformingOptimizer,
        ... ClampedGradientOptimizer,
        ... TopQGradientOptimizer,
        ... )
        >>> import torch as tr
        >>> from functools import partial

        >>> x1 = tr.ones(3, requires_grad=True)  # shape-(3,)

        Our optimizer will retain only the top-33rd percentile elements in the gradient:
        the smallest elements will be zero'd. Then the resulting gradient will be
        clamped so that its largest possible entry is `2.8`. Finally, the standard `SGD`
        optimizer will be used, with `lr=1.0`, to update the parameter(s) using the
        transformed gradients.

        We specify `TopQGradientOptimizer` and then `ClampedGradientOptimizer`; the
        transformations are applied in order from left to right. Providing per-optimizer
        defaults is achieved most naturally using :py:func:`functools.partial`.

        >>> optim = ChainedParamTransformingOptimizer(
        ...     partial(TopQGradientOptimizer, q=0.33),
        ...     partial(ClampedGradientOptimizer, clamp_max=2.8),
        ...     params=[x1],
        ...     lr=1.0,
        ...     param_ndim=None, # we don't want any broadcasting to occur
        ... )
        ClampedGradientOptimizer ○ TopQGradientOptimizer [SGD](
        Parameter Group 0
            clamp_max: 2.8
            clamp_min: None
            dampening: 0
            dq: 0.0
            grad_bias: 0
            grad_scale: 1
            lr: 1.0
            maximize: False
            momentum: 0
            nesterov: False
            param_ndim: None
            q: 0.33
            weight_decay: 0
        )

        Let's verify that `optim` transforms our gradients as-expected.

        >>> (tr.tensor([1.0, 2.0, 3.0]) * x1).sum().backward()
        >>> optim.step()
        >>> x1.grad  # element-0 should be zero'd by top-q; element-2 should be clamped to 2.8
        tensor([0.0000, 2.0000, 2.8000])

        See that `SGD([x1], lr=1.0).step()` is used to update our parameters; this can be controlled via the `InnerOpt` argument.

        >>> x1
        tensor([ 1.0000, -1.0000, -1.8000], requires_grad=True)

        **Adding Parameter Groups**

        Our chained gradient-transforming optimizers mirror their states with `optim`
        and `SGD`, thus we can add parameter groups and the group's settings will be
        applied to our chain as-expected.

        Let's add a 2D parameter, where we want to apply the top-q sparsification
        row-wise (via `param_ndim=1`), and retain only 64th-percentile gradient
        elements.

        >>> x2 = tr.ones(2, 3, requires_grad=True)  # shape-(2, 3)
        >>> optim.add_param_group(dict(params=x2, param_ndim=1, q=0.64))
        >>> optim
        ClampedGradientOptimizer ○ TopQGradientOptim [SGD](
        Parameter Group 0
            clamp_max: 2.8
            clamp_min: None
            dampening: 0
            dq: 0.0
            grad_bias: 0
            grad_scale: 1
            lr: 1.0
            maximize: False
            momentum: 0
            nesterov: False
            param_ndim: None
            q: 0.33
            weight_decay: 0
        Parameter Group 1
            clamp_max: 2.8
            clamp_min: None
            dampening: 0
            dq: 0.0
            grad_bias: 0
            grad_scale: 1
            lr: 1.0
            maximize: False
            momentum: 0
            nesterov: False
            param_ndim: 1
            q: 0.64

        >>> optim.zero_grad()
        >>> (tr.tensor([1.0, 2.0, 3.0]) * (x1 + x2)).sum().backward()
        >>> optim.step()
        >>> x1.grad
        tensor([0.0000, 2.8000, 2.8000])
        >>> x2.grad
        tensor([[0.0000, 0.0000, 2.8000],
            [0.0000, 0.0000, 2.8000]])
        """
        self._chain = ()

        super().__init__(
            params,
            InnerOpt,
            param_ndim=param_ndim,
            grad_scale=grad_scale,
            grad_bias=grad_bias,
            defaults=defaults,
            **inner_opt_kwargs,
        )
        for _opt in transforming_optimizers:
            if not instantiates_to(_opt, ParamTransformingOptimizer):
                raise TypeError(
                    f"*transforming_optimizers must contain `Type[ParamTransformingOptimizer]`, got: {transforming_optimizers}"
                )
        self._chain = tuple(
            opt(params, InnerOpt=self.inner_opt, defaults=self.defaults)
            for opt in transforming_optimizers
        )

    def _pre_step_transform_(self, param: Tensor, optim_group: DatumParamGroup) -> None:
        # [f1, f2, f3] -> f3(f2(f1(param)))
        for opt in self._chain:
            opt._pre_step_transform_(param=param, optim_group=optim_group)

    def _post_step_transform_(
        self, param: Tensor, optim_group: DatumParamGroup
    ) -> None:
        # [f1, f2, f3] -> f3(f2(f1(param)))
        for opt in self._chain:
            opt._post_step_transform_(param=param, optim_group=optim_group)

    def __setstate__(self, state: dict):
        # synchornize state between `self`, members of `self._chain`,
        # and `self.inner_opt`
        self.inner_opt.__setstate__(state)
        state = self.inner_opt.__getstate__()  # type: ignore
        for c in self._chain:
            c.__setstate__(state)
        super().__setstate__(state)

    def __repr__(self) -> str:
        return (
            super()
            .__repr__()
            .replace(
                type(self).__name__,
                " ○ ".join(type(c).__name__ for c in self._chain[::-1]),
            )
        )



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from abc import ABCMeta, abstractmethod
from functools import wraps
from typing import Callable, Dict, Iterable, TypeVar, Union, cast
from weakref import WeakSet

import torch as tr

from .itertools import flatten_params

T = TypeVar("T", bound=Callable)
NoneType = type(None)


def freeze(
    *items: Union[
        tr.Tensor,
        tr.nn.Module,
        tr.optim.Optimizer,
        Iterable[tr.Tensor],
        Iterable[Dict[str, Iterable[tr.Tensor]]],
    ]
) -> Callable[[], None]:
    """'Freezes' collections of tensors by setting `requires_grad=False`.
    Returns a callable that, when called, restores the state of the tensors.

    Parameters
    ----------
    *items: tr.Tensor | tr.nn.Module | tr.optim.Optimizer | Iterable[tr.Tensor] | Iterable[Dict[str, Iterable[tr.Tensor]]]
        Tensors, modules, optimizers, or param-groups. All tensors/parameters must
        be leaf tensors [1]_ .

    Returns
    -------
    unfreeze : Callable[[], None]
        Can be called without any input to restore the states of the frozen tensors.

    Notes
    -----
    'Unfreezing' the tensors restores their original states faithfully.

    References
    ----------
    .. [1]  https://pytorch.org/docs/stable/generated/torch.Tensor.is_leaf.html

    Examples
    --------
    >>> import torch as tr
    >>> from rai_toolbox.utils import freeze

    Basic behavior

    >>> x = tr.tensor(1.0, requires_grad=True)
    >>> unfreeze = freeze(x)
    >>> x.requires_grad
    False
    >>> unfreeze()
    >>> x.requires_grad
    True

    Freezing a module

    >>> from torch.nn import Linear
    >>> m = Linear(2, 3)
    >>> m.weight.requires_grad, m.bias.requires_grad
    (True, True)
    >>> unfreeze = freeze(m)
    >>> m.weight.requires_grad, m.bias.requires_grad
    (False, False)
    >>> unfreeze()
    >>> m.weight.requires_grad, m.bias.requires_grad
    (True, True)
    """
    seen = {True: WeakSet(), False: WeakSet()}
    for item in items:
        if isinstance(item, tr.nn.Module):
            item = item.parameters()
        elif isinstance(item, tr.optim.Optimizer):
            item = item.param_groups

        for param in flatten_params(item):
            seen[param.requires_grad].add(param)

    for param in seen[True]:
        param.requires_grad_(False)

    def restore_state():
        for item in (True, False):
            for p in seen[item]:
                p.requires_grad_(item)

    return restore_state


class ContextDecorator(metaclass=ABCMeta):
    @abstractmethod
    def __enter__(self):  # pragma: no cover
        raise NotImplementedError()

    @abstractmethod
    def __exit__(self, type, value, traceback):  # pragma: no cover
        raise NotImplementedError()

    def __call__(self, func: T) -> T:
        @wraps(func)
        def wrapper(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return cast(T, wrapper)


class frozen(ContextDecorator):
    """A context manager/decorator for 'freezing' collections of tensors; i.e.
    `requires_grad` is set to `False` for the tensors during the context."""

    def __init__(
        self,
        *items: Union[
            tr.Tensor,
            tr.nn.Module,
            tr.optim.Optimizer,
            Iterable[tr.Tensor],
            Iterable[Dict[str, Iterable[tr.Tensor]]],
        ],
    ) -> None:
        """
        Parameters
        ----------
        *items: tr.Tensor | tr.nn.Module | tr.optim.Optimizer | Iterable[tr.Tensor] | Iterable[Dict[str, Iterable[tr.Tensor]]]
            Tensors, modules, optimizers, or param-groups to be frozen. All tensors/
            parameters must be leaf tensors [1]_ .

        References
        ----------
        .. [1]  https://pytorch.org/docs/stable/generated/torch.Tensor.is_leaf.html

        Examples
        --------
        >>> import torch as tr
        >>> from rai_toolbox.utils._implementations import frozen

        Demonstrating `frozen` as a context manager.

        >>> x = tr.tensor(1.0, requires_grad=True)
        >>> with frozen(x):
        ...    print(x.requires_grad)
        False
        >>> x.requires_grad
        True

        Demonstrating `frozen` as a decorator.

        >>> x = tr.tensor(1.0, requires_grad=True)
        >>> @frozen(x)
        ... def f():
        ...     print("hello world")
        ...     return x.requires_grad
        >>> x.requires_grad # x isn't frozen until f is called
        True
        >>> f()
        hello world
        False
        >>> x.requires_grad
        True
        """
        self._items = items

    def __enter__(self) -> None:
        self._unfreeze = freeze(*self._items)

    def __exit__(self, type, value, traceback) -> None:
        self._unfreeze()


class evaluating(ContextDecorator):
    """A context manager / decorator that temporarily places one
    or more modules in eval mode during the context."""

    def __init__(self, *modules: tr.nn.Module) -> None:
        """
        Parameters
        ----------
        *modules: tr.nn.Module

        Notes
        -----
        A module's state is restored faithfully; e.g., a module that
        was already in eval mode will not be placed in train mode upon
        leaving the `evaluating` context.

        Examples
        --------
        >>> from torch.nn import Linear
        >>> from rai_toolbox import evaluating

        Using `evaluating` as a context manager.

        >>> module = Linear(1, 1)
        >>> module.training
        True
        >>> with evaluating(module):
        ...     print(module.training)
        False
        >>> module.training
        True

        Using `evaluating` as a decorator.

        >>> def f():
        ...     print("hello world")
        ...     return module.training
        >>> f = evaluating(module)(f)
        >>> module.training
        True
        >>> f()
        hello world
        False
        >>> module.training
        True
        """
        self._states: Dict[bool, WeakSet[tr.nn.Module]] = {
            True: WeakSet(),
            False: WeakSet(),
        }

        self._states[True].update(m for m in modules if m.training)
        self._states[False].update(m for m in modules if not m.training)

    def __enter__(self) -> None:
        for train_status in self._states:
            for m in self._states[train_status]:
                m.eval()

    def __exit__(self, type, value, traceback) -> None:
        for train_status in self._states:
            for module in self._states[train_status]:
                module.train(train_status)



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from ._jensen_shannon_divergence import jensen_shannon_divergence

__all__ = ["jensen_shannon_divergence"]



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from ._imagenet_base import ImageNet, ImageNetM10, RestrictedImageNet
from .cifar10_extensions import CIFAR10P1
from .cifar_corruptions import CIFAR10C, CIFAR100C

__all__ = [
    "CIFAR10C",
    "CIFAR100C",
    "CIFAR10P1",
    "ImageNet",
    "ImageNetM10",
    "RestrictedImageNet",
]



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
from numbers import Real
from typing import Any, Iterable, Mapping, Optional, Tuple, TypeVar, Union, cast

import torch as tr

T = TypeVar("T", bound=Any)


class Unsatisfiable(AssertionError):  # pragma: no cover
    pass


def get_device(obj: Union[tr.nn.Module, tr.Tensor]) -> tr.device:
    if isinstance(obj, tr.nn.Module):
        for p in obj.parameters():
            return p.device
        return tr.device("cpu")

    elif isinstance(obj, tr.Tensor):
        return obj.device

    else:  # pragma: no cover
        raise TypeError(f"Expected torch.nn.Module or torch.Tensor, got {obj}")


def _safe_name(x: Any) -> str:
    return getattr(x, "__name__", str(x))


def value_check(
    name: str,
    value: T,
    *,
    type_: Union[type, Tuple[type, ...]] = Real,
    min_: Optional[Union[int, float]] = None,
    max_: Optional[Union[int, float]] = None,
    incl_min: bool = True,
    incl_max: bool = True,
    optional: bool = False,
    lower_name: str = "",
    upper_name: str = "",
) -> T:
    """
    For internal use only.

    Used to check the type of `value`. Numerical types can also be bound-checked.

    Examples
    --------
    >>> value_check("x", 1, type_=str)
    TypeError: `x` must be of type(s) `str`, got 1 (type: int)

    >>> value_check("x", 1, min_=20)
    ValueError: `x` must satisfy 20 <= x  Got: 1

    >>> value_check("x", 1, min_=1, incl_min=False)
    ValueError: `x` must satisfy 1 < x  Got: 1

    >>> value_check("x", 1, min_=1, incl_min=True) # ok
    1
    >>> value_check("x", 0.0, min_=-10, max_=10)  # ok
    0.0

    Raises
    ------
    TypeError, ValueError"""
    # check internal params
    assert isinstance(name, str), name
    assert min_ is None or isinstance(min_, (int, float)), min_
    assert max_ is None or isinstance(max_, (int, float)), max_
    assert isinstance(incl_min, bool), incl_min
    assert isinstance(incl_max, bool), incl_max

    if optional and value is None:
        return value

    if not isinstance(value, type_):
        raise TypeError(
            f"`{name}` must be {'None or' if optional else ''}of type(s) `{_safe_name(type_)}`, got {value} (type: {_safe_name(type(value))})"
        )

    if min_ is not None and max_ is not None:
        if incl_max and incl_min:
            if not (min_ <= max_):
                raise Unsatisfiable(f"{min_} <= {max_}")
        elif not min_ < max_:
            raise Unsatisfiable(f"{min_} < {max_}")

    min_satisfied = (
        (min_ <= value if incl_min else min_ < value) if min_ is not None else True
    )
    max_satisfied = (
        (value <= max_ if incl_min else value < max_) if max_ is not None else True
    )

    if not min_satisfied or not max_satisfied:
        lsymb = "<=" if incl_min else "<"
        rsymb = "<=" if incl_max else "<"

        err_msg = f"`{name}` must satisfy"

        if min_ is not None:
            if lower_name:  # pragma: no cover
                min_ = f"{lower_name}(= {min_})"  # type: ignore
            err_msg += f" {min_} {lsymb}"

        err_msg += f" {name}"

        if max_ is not None:
            if upper_name:
                max_ = f"{upper_name}(= {max_})"  # type: ignore
            err_msg += f" {rsymb} {max_}"

        err_msg += f"  Got: {value}"

        raise ValueError(err_msg)
    return cast(T, value)


def check_param_group_value(
    name: str,
    param_groups: Iterable[Mapping[str, Any]],
    *,
    type_: Union[type, Tuple[type, ...]] = Real,
    min_: Optional[Union[int, float]] = None,
    max_: Optional[Union[int, float]] = None,
    incl_min: bool = True,
    incl_max: bool = True,
    optional: bool = False,
) -> None:
    for group in param_groups:
        value_check(
            name,
            group[name],
            type_=type_,
            max_=max_,
            min_=min_,
            incl_min=incl_min,
            incl_max=incl_max,
            optional=optional,
        )


def _reshape_to_batch(x: tr.Tensor, param_ndim: Optional[int]) -> tr.Tensor:
    """Reshapes to a shape-`(N, d1, ..., dm)`, where `(d1, ..., dm)` has `param_ndim`
    dimensions. Dimensions will be added or consolidated to achieve this."""
    if param_ndim is None:
        param_ndim = x.ndim

    if param_ndim < 0:
        param_ndim += x.ndim

    # `1 + param_ndim` is the required dimensionality
    # for a shape-(N, d0, d1, ...) tensor, where (d0, d1, ...)
    # is the shape of the param_ndim-dimension tensor.
    #
    # We compute `ndim_delta` to determine if we need to add
    # or consolidate dimensions to create the shape-(N, d0, d1, ...)
    # tensor.
    ndim_delta = (1 + param_ndim) - x.ndim

    if ndim_delta > 0:
        # E.g.:
        #   p.shape: (d0, )
        #   desired shape: (N=1, d0)
        x = x[ndim_delta * (None,)]
    elif ndim_delta < 0:
        # E.g.:
        #   p.shape: (d0, d1, d2, d3)
        #   desired shape: (N=d0*d1, d2, d3)
        x = x.view(-1, *x.shape[x.ndim - param_ndim :])
    if x.ndim < 2:  # make at least 2D
        # (N,) -> (N, 1)
        x = x.view(*x.shape, *(1,) * (2 - x.ndim))
    return x


def to_batch(p: tr.Tensor, param_ndim: Optional[int]) -> tr.Tensor:
    """
    Returns a view of `p`, reshaped as shape-(N, d0, ...) where (d0, ...)
    has `param_ndim` entries.

    See Parameters for further description

    Parameters
    ----------
    p : Tensor

    param_ndim: Optional[int]
        Determines the shape of the resulting parameter

        - A positive number determines the dimensionality of the tensor that the transformation will act on.
        - A negative number indicates the 'offset' from the dimensionality of the tensor.
        - `None` means that the transformation will be applied to the tensor without any broadcasting.

    Returns
    -------
    reshaped_p: Tensor, shape-(N, d0, ...)
        Where
        - (d0, ...) is of length `param_ndim` for `param_ndim > 0`
        - (d0, ...) is (1,) for `param_ndim == 0`
        - (d0, ...) is of length `p.ndim - |param_ndim|` for `param_ndim < 0`

    Examples
    --------
    >>> import torch as tr
    >>> x = tr.rand((3, 5, 2))

    >>> to_batch(x, param_ndim=0).shape
    torch.Size([30, 1])

    >>> to_batch(x, param_ndim=1).shape
    torch.Size([15, 2])

    >>> to_batch(x, param_ndim=2).shape
    torch.Size([3, 5, 2])

    >>> to_batch(x, param_ndim=3).shape
    torch.Size([1, 3, 5, 2])

    >>> to_batch(x, param_ndim=None).shape  # same as `param_ndim=x.ndim`
    torch.Size([1, 3, 5, 2])

    >>> to_batch(x, param_ndim=-1).shape
    torch.Size([3, 5, 2])

    >>> to_batch(x, param_ndim=-2).shape
    torch.Size([15, 2])

    >>> to_batch(x, param_ndim=-3).shape
    torch.Size([30, 1])
    """

    # atleast_2d needed for case where p was scalar
    vp = _reshape_to_batch(p, param_ndim=param_ndim)

    if p.grad is not None:
        vp.grad = _reshape_to_batch(p.grad, param_ndim=param_ndim)

    # vp (vp.grad) must be a view of p (p.grad). There is
    # not a simple way to assert this.

    # our views must be size-preserving
    assert tr.numel(vp) == tr.numel(p)
    return vp


def validate_param_ndim(param_ndim: Optional[int], p: tr.Tensor) -> None:
    if param_ndim is not None and p.ndim < abs(param_ndim):
        raise ValueError(
            f"`param_ndim={param_ndim}` specified for parameter "
            f"with ndim={p.ndim} is not valid. `abs(param_ndim) <= "
            f"ndim` must hold."
        )



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from typing import Union

import torch
from torch.optim import Optimizer

from rai_toolbox._typing import OptimParams
from rai_toolbox._utils import check_param_group_value, value_check

from .lp_space import L1qNormedGradientOptim, L2NormedGradientOptim, SignedGradientOptim
from .optimizer import ParamTransformingOptimizer

_TINY = torch.finfo(torch.float32).tiny

__all__ = [
    "FrankWolfe",
    "L1FrankWolfe",
    "L2FrankWolfe",
    "L1qFrankWolfe",
    "L1qNormedGradientOptim",
]


class FrankWolfe(Optimizer):
    r"""Implements the Frank-Wolfe minimization algorithm [1]_.

    .. math::
        w_{k+1} = (1 - l_r) w_k + l_r * s_k

    where :math:`s_k` is the linear minimization oracle (LMO).

    It is critical to note that this optimizer assumes that the `.grad` attribute
    of each parameter has been modified so as to store the *negative* of the LMO
    for that parameter, and not the gradient itself.


    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Frank%E2%80%93Wolfe_algorithm#Algorithm"""

    def __init__(
        self,
        params: OptimParams,
        *,
        lr: float = 2.0,
        lmo_scaling_factor: float = 1.0,
        use_default_lr_schedule: bool = True,
        div_by_zero_eps: float = torch.finfo(torch.float32).tiny,
    ):
        r"""
        Parameters
        ----------
        params : Iterable
            Iterable of tensor parameters to optimize or dicts defining parameter groups.

        lr :  float, optional (default=2.0)
            Indicates the weight with which the LMO contributes to the parameter
            update. See ``use_default_lr_schedule`` for additional details. If
            ``use_default_lr_schedule=False`` then ``lr`` must be be in the
            domain [0, 1].

        lmo_scaling_factor : float, optional (default=1.0)
            A scaling factor applied to :math:`s_k` prior to each step.

        use_default_lr_schedule : bool, optional (default=True)
            If ``True``, then the per-parameter "learning rate" is scaled
            by :math:`\hat{l_r} = l_r / (l_r + k)` where k is the update index
            for that parameter.

        div_by_zero_eps : float, optional (default=`torch.finfo(torch.float32).tiny`)
            Prevents div-by-zero error in learning rate schedule.
        """
        self._eps = value_check("div_by_zero_eps", div_by_zero_eps, min_=0.0)
        self._use_default_lr_schedule = value_check(
            "use_default_lr_schedule", use_default_lr_schedule, type_=bool
        )

        defaults = dict(lr=lr, lmo_scaling_factor=lmo_scaling_factor)

        super().__init__(params, defaults)  # type: ignore

        check_param_group_value("lmo_scaling_factor", self.param_groups)

        check_param_group_value(
            "lr",
            self.param_groups,
            min_=0.0,
            max_=1.0 if not self._use_default_lr_schedule else None,
        )

    @torch.no_grad()
    def step(self, closure=None):
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        for group in self.param_groups:
            lr = group["lr"]
            lmo_scale = group["lmo_scaling_factor"]
            for p in group["params"]:
                if p.grad is not None:
                    state = self.state[p]
                    if not state:
                        state["step"] = 0

                    alpha = (
                        lr / max(self._eps, lr + state["step"])
                        if self._use_default_lr_schedule
                        else lr
                    )

                    # alpha weighting included in LMO (more efficient)
                    weighted_lmo = -(alpha * lmo_scale) * p.grad

                    p *= 1 - alpha
                    p += weighted_lmo

                    state["step"] += 1

        return loss


class L1qFrankWolfe(L1qNormedGradientOptim):
    r"""A Frank-Wolfe [1]_ optimizer that, when computing the LMO, sparsifies a
    parameter's gradient. Each updated parameter is constrained to fall within an
    :math:`\epsilon`-sized ball in :math:`L^1` space, centered on the origin.

    The sparsification process retains only the signs (i.e., :math:`\pm 1`) of the
    gradient's elements. The transformation is applied to the gradient in accordance
    with `param_ndim`.

    See Also
    --------
    FrankWolfe
    L1FrankWolfe
    L2FrankWolfe
    LinfFrankWolfe

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Frank%E2%80%93Wolfe_algorithm#Algorithm
    """

    def __init__(
        self,
        params: OptimParams,
        *,
        q: float,
        epsilon: float,
        dq: float = 0.0,
        lr: float = 2.0,
        use_default_lr_schedule: bool = True,
        param_ndim: Union[int, None] = -1,
        div_by_zero_eps: float = _TINY,
        generator: torch.Generator = torch.default_generator,
    ):
        r"""
        Parameters
        ----------
        params : Sequence[Tensor] | Iterable[Mapping[str, Any]]
            Iterable of parameters or dicts defining parameter groups.

        q : float
            Specifies the (fractional) percentile of absolute-largest gradient elements
            to retain when sparsifying the gradient. E.g., `q=0.9`means that only the
            gradient elements within the 90th-percentile will be retained.

            Must be within `[0.0, 1.0]`. The sparsification is applied to the gradient
            in accordance to `param_ndim`.

        epsilon : float
            Specifies the size of the L1-space ball that all parameters will be
            projected into, post optimization step.

        dq : float, optional (default=0.0)
            If specified, the sparsity factor for each gradient transformation will
            be drawn from a uniform distribution over :math:`[q - dq, q + dq] \in [0.0, 1.0]`.

        lr :  float, optional (default=2.0)
            Indicates the weight with which the LMO contributes to the parameter
            update. See `use_default_lr_schedule` for additional details. If
            `use_default_lr_schedule=False` then `lr` must be be in the
            domain `[0, 1]`.

        use_default_lr_schedule : bool, optional (default=True)
            If `True`, then the per-parameter "learning rate" is scaled
            by :math:`\hat{l_r} = l_r / (l_r + k)` where k is the update index
            for that parameter, which starts at 0.

        param_ndim : Union[int, None], optional (default=-1)
            Determines how a parameter and its gradient is temporarily reshaped prior
            to being passed to both `_pre_step_transform_` and `_post_step_transform_`.
            By default,the transformation broadcasts over the tensor's first dimension
            in a batch-like style. This can be specified per param-group

            - A positive number determines the dimensionality of the tensor that the transformation will act on.
            - A negative number indicates the 'offset' from the dimensionality of the tensor (see "Notes" for examples).
            - `None` means that the transformation will be applied directly to the tensor without any broadcasting.

            See `ParamTransformingOptimizer` for more details and examples.

        defaults : Optional[Dict[str, Any]]
            Specifies default parameters for all parameter groups.

        div_by_zero_eps : float, optional (default=`torch.finfo(torch.float32).tiny`)
            Prevents div-by-zero error in learning rate schedule.

        generator : torch.Generator, optional (default=`torch.default_generator`)
            Controls the RNG source.

        Examples
        --------
        Using `L1qFrankWolfe`, we'll sparsify the parameter's gradient to retain signs of
        the top 70% elements, and we'll constrain the updated parameter to fall within a
        :math:`L^1`-ball of radius `1.8`.

        >>> import torch as tr
        >>> from rai_toolbox.optim import L1qFrankWolfe

        Creating a parameter for our optimizer to update, and our optimizer. We
        specify `param_ndim=None` so that the sparsification/normalization occurs on the
        gradient without any broadcasting.

        >>> x = tr.tensor([1.0, 1.0, 1.0], requires_grad=True)
        >>> optim = L1qFrankWolfe(
        ...     [x],
        ...     q=0.30,
        ...     epsilon=1.8,
        ...     param_ndim=None,
        ... )

        Performing a simple calculation with `x` and performing backprop to create
        a gradient.

        >>> (tr.tensor([0.0, 1.0, 2.0]) * x).sum().backward()

        Performing a step with our optimizer uses the Frank-Wolfe algorithm to update
        its parameters; the resulting parameter was updated with a LMO based on a
        sparsified, sign-only gradient. Note that the parameter falls within/on the
        :math:`L^1`-ball of radius `1.8`.

        >>> optim.step()
        >>> x  # the updated parameter; has a L1-norm of 1.8
        tensor([ 0.0000, -0.9000, -0.9000], requires_grad=True)
        """
        super().__init__(
            params,
            InnerOpt=FrankWolfe,
            lmo_scaling_factor=epsilon,
            q=q,
            dq=dq,
            param_ndim=param_ndim,
            div_by_zero_eps=div_by_zero_eps,
            generator=generator,
            lr=lr,
            use_default_lr_schedule=use_default_lr_schedule,
        )


class L1FrankWolfe(ParamTransformingOptimizer):
    r"""A Frank-Wolfe [1]_ optimizer that constrains each updated parameter to fall
    within an :math:`\epsilon`-sized ball in :math:`L^1` space, centered on the origin.

    Notes
    -----
    The method `L1NormedGradientOptim._pre_step_transform_` is responsible for
    computing the *negative* linear minimization oracle for a parameter and storing it
    on `param.grad`.

    See Also
    --------
    FrankWolfe
    L2FrankWolfe
    LinfFrankWolfe
    L1qFrankWolfe

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Frank%E2%80%93Wolfe_algorithm#Algorithm
    """

    def __init__(
        self,
        params: OptimParams,
        *,
        epsilon: float,
        lr: float = 2.0,
        use_default_lr_schedule: bool = True,
        param_ndim: Union[int, None] = -1,
        div_by_zero_eps: float = _TINY,
    ):
        r"""
        Parameters
        ----------
        params : Sequence[Tensor] | Iterable[Mapping[str, Any]]
            Iterable of parameters or dicts defining parameter groups.

        epsilon : float
            The radius of the of the L1 ball to which each updated parameter will be
            constrained. Can be specified per parameter-group.

        lr :  float, optional (default=2.0)
            Indicates the weight with which the LMO contributes to the parameter
            update. See `use_default_lr_schedule` for additional details. If
            `use_default_lr_schedule=False` then `lr` must be be in the
            domain `[0, 1]`.

        use_default_lr_schedule : bool, optional (default=True)
            If ``True``, then the per-parameter "learning rate" is scaled
            by :math:`\hat{l_r} = l_r / (l_r + k)` where k is the update index
            for that parameter, which starts at 0.

        param_ndim : Union[int, None], optional (default=-1)
            Determines how a parameter and its gradient is temporarily reshaped prior
            to being passed to both `_pre_step_transform_` and `_post_step_transform_`.
            By default,the transformation broadcasts over the tensor's first dimension
            in a batch-like style. This can be specified per param-group

            - A positive number determines the dimensionality of the tensor that the transformation will act on.
            - A negative number indicates the 'offset' from the dimensionality of the tensor (see "Notes" for examples).
            - `None` means that the transformation will be applied directly to the tensor without any broadcasting.

            See `ParamTransformingOptimizer` for more details and examples.

        div_by_zero_eps : float, optional (default=`torch.finfo(torch.float32).tiny`)
            Prevents div-by-zero error in learning rate schedule.

        Examples
        --------
        Using `L1FrankWolfe`, we'll constrain the updated parameter to fall within a
        :math:`L^1`-ball of radius `1.8`.

        >>> import torch as tr
        >>> from rai_toolbox.optim import L1FrankWolfe

        Creating a parameter for our optimizer to update, and our optimizer. We
        specify `param_ndim=None` so that the constrain occurs on the parameter without
        any broadcasting.

        >>> x = tr.tensor([1.0, 1.0], requires_grad=True)
        >>> optim = L1FrankWolfe([x], epsilon=1.8, param_ndim=None)

        Performing a simple calculation with `x` and performing backprop to create
        a gradient.

        >>> (tr.tensor([1.0, 2.0]) * x).sum().backward()

        Performing a step with our optimizer uses the Frank-Wolfe algorithm to update
        its parameters. Note that the updated parameter falls within/on the
        :math:`L^1`-ball of radius `1.8`.

        >>> optim.step()
        >>> x
        tensor([ 0.0000, -1.8000], requires_grad=True)
        """
        super().__init__(
            params,
            InnerOpt=FrankWolfe,
            lmo_scaling_factor=epsilon,
            lr=lr,
            use_default_lr_schedule=use_default_lr_schedule,
            param_ndim=param_ndim,
            div_by_zero_eps=div_by_zero_eps,
        )

    def _pre_step_transform_(self, param: torch.Tensor, **_unused_kwargs) -> None:
        if param.grad is None:  # pragma: no cover
            return
        # Computes the negative linear minimization oracle and sets it to
        # `param.grad`
        argmax = torch.argmax(torch.abs(param.grad), dim=1)
        signs = torch.sign(param.grad[torch.arange(len(param)), argmax])
        param.grad.mul_(0.0)
        param.grad[torch.arange(len(param)), argmax] = signs


class L2FrankWolfe(L2NormedGradientOptim):
    r"""A Frank-Wolfe [1]_ optimizer that constrains each updated parameter to fall
    within an :math:`\epsilon`-sized ball in :math:`L^2` space, centered on the origin.

    Notes
    -----
    The method `L2NormedGradientOptim._pre_step_transform_` is responsible for
    computing the *negative* linear minimization oracle for a parameter and storing it
    on `param.grad`.

    See Also
    --------
    FrankWolfe
    L1FrankWolfe
    LinfFrankWolfe
    L1qFrankWolfe

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Frank%E2%80%93Wolfe_algorithm#Algorithm
    """

    def __init__(
        self,
        params: OptimParams,
        *,
        epsilon: float,
        lr: float = 2.0,
        use_default_lr_schedule: bool = True,
        param_ndim: Union[int, None] = -1,
        div_by_zero_eps: float = _TINY,
    ):
        r"""
        Parameters
        ----------
        params : Sequence[Tensor] | Iterable[Mapping[str, Any]]
            Iterable of parameters or dicts defining parameter groups.

        epsilon : float
            The radius of the of the L2 ball to which each updated parameter will be
            constrained. Can be specified per parameter-group.

        lr :  float, optional (default=2.0)
            Indicates the weight with which the LMO contributes to the parameter
            update. See `use_default_lr_schedule` for additional details. If
            `use_default_lr_schedule=False` then `lr` must be be in the
            domain `[0, 1]`.

        use_default_lr_schedule : bool, optional (default=True)
            If ``True``, then the per-parameter "learning rate" is scaled
            by :math:`\hat{l_r} = l_r / (l_r + k)` where k is the update index
            for that parameter, which starts at 0.

        param_ndim : Union[int, None], optional (default=-1)
            Determines how a parameter and its gradient is temporarily reshaped prior
            to being passed to both `_pre_step_transform_` and `_post_step_transform_`.
            By default,the transformation broadcasts over the tensor's first dimension
            in a batch-like style. This can be specified per param-group

            - A positive number determines the dimensionality of the tensor that the transformation will act on.
            - A negative number indicates the 'offset' from the dimensionality of the tensor (see "Notes" for examples).
            - `None` means that the transformation will be applied directly to the tensor without any broadcasting.

            See `ParamTransformingOptimizer` for more details and examples.

        div_by_zero_eps : float, optional (default=`torch.finfo(torch.float32).tiny`)
            Prevents div-by-zero error in learning rate schedule.

        Examples
        --------
        Using `L2FrankWolfe`, we'll constrain the updated parameter to fall within a
        :math:`L^2`-ball of radius `1.8`.

        >>> import torch as tr
        >>> from rai_toolbox.optim import L2FrankWolfe

        Creating a parameter for our optimizer to update, and our optimizer. We
        specify `param_ndim=None` so that the constrain occurs on the parameter without any
        broadcasting.

        >>> x = tr.tensor([1.0, 1.0], requires_grad=True)
        >>> optim = L2FrankWolfe([x], epsilon=1.8, param_ndim=None)

        Performing a simple calculation with `x` and performing backprop to create
        a gradient.

        >>> (tr.tensor([1.0, 2.0]) * x).sum().backward()

        Performing a step with our optimizer uses the Frank-Wolfe algorithm to update
        its parameters. Note that the updated parameter falls within/on the
        :math:`L^2`-ball of radius `1.8`.

        >>> optim.step()
        >>> x
        tensor([-0.8050, -1.6100], requires_grad=True)
        """
        super().__init__(
            params,
            InnerOpt=FrankWolfe,
            lmo_scaling_factor=epsilon,
            lr=lr,
            use_default_lr_schedule=use_default_lr_schedule,
            param_ndim=param_ndim,
            div_by_zero_eps=div_by_zero_eps,
        )


class LinfFrankWolfe(SignedGradientOptim):
    r"""A Frank-Wolfe [1]_ optimizer that constrains each updated parameter to fall
    within an :math:`\epsilon`-sized ball in :math:`L^\infty` space, centered on the origin.

    Notes
    -----
    The method `SignedGradientOptim._pre_step_transform_` is responsible for
    computing the *negative* linear minimization oracle for a parameter and storing it
    on `param.grad`.

    See Also
    --------
    FrankWolfe
    L1FrankWolfe
    L2FrankWolfe
    L1qFrankWolfe

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Frank%E2%80%93Wolfe_algorithm#Algorithm

    Examples
    --------
    Using `LinfFrankWolfe`, we'll constrain the updated parameter to fall within a
    :math:`L^\infty`-ball of radius `1.8`.

    >>> import torch as tr
    >>> from rai_toolbox.optim import LinfFrankWolfe

    Creating a parameter for our optimizer to update, and our optimizer. We
    specify `param_ndim=None` so that the constrain occurs on the parameter without any
    broadcasting.

    >>> x = tr.tensor([1.0, 1.0], requires_grad=True)
    >>> optim = L2FrankWolfe([x], epsilon=1.8, param_ndim=None)

    Performing a simple calculation with `x` and performing backprop to create
    a gradient.

    >>> (tr.tensor([1.0, 2.0]) * x).sum().backward()

    Performing a step with our optimizer uses the Frank-Wolfe algorithm to update
    its parameters. Note that the updated parameter falls within/on the
    :math:`L^\infty`-ball of radius `1.8`.

    >>> optim.step()
    >>> x
    tensor([-1.8000, -1.8000], requires_grad=True)"""

    def __init__(
        self,
        params: OptimParams,
        *,
        epsilon: float,
        lr: float = 2.0,
        use_default_lr_schedule: bool = True,
        param_ndim: Union[int, None] = -1,
        div_by_zero_eps: float = _TINY,
    ):
        r"""
        Parameters
        ----------
        params : Sequence[Tensor] | Iterable[Mapping[str, Any]]
            Iterable of parameters or dicts defining parameter groups.

        epsilon : float
            The radius of the of the L-inf ball to which each updated parameter will be
            constrained. Can be specified per parameter-group.

        lr :  float, optional (default=2.0)
            Indicates the weight with which the LMO contributes to the parameter
            update. See `use_default_lr_schedule` for additional details. If
            `use_default_lr_schedule=False` then `lr` must be be in the
            domain `[0, 1]`.

        use_default_lr_schedule : bool, optional (default=True)
            If ``True``, then the per-parameter "learning rate" is scaled
            by :math:`\hat{l_r} = l_r / (l_r + k)` where k is the update index
            for that parameter, which starts at 0.

        param_ndim : Union[int, None], optional (default=-1)
            Determines how a parameter and its gradient is temporarily reshaped prior
            to being passed to both `_pre_step_transform_` and `_post_step_transform_`.
            By default,the transformation broadcasts over the tensor's first dimension
            in a batch-like style. This can be specified per param-group

            - A positive number determines the dimensionality of the tensor that the transformation will act on.
            - A negative number indicates the 'offset' from the dimensionality of the tensor (see "Notes" for examples).
            - `None` means that the transformation will be applied directly to the tensor without any broadcasting.

            See `ParamTransformingOptimizer` for more details and examples.

        div_by_zero_eps : float, optional (default=`torch.finfo(torch.float32).tiny`)
            Prevents div-by-zero error in learning rate schedule.
        """
        super().__init__(
            params,
            InnerOpt=FrankWolfe,
            lmo_scaling_factor=epsilon,
            lr=lr,
            use_default_lr_schedule=use_default_lr_schedule,
            param_ndim=param_ndim,
            div_by_zero_eps=div_by_zero_eps,
        )



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from numbers import Integral
from typing import Any, Callable, Optional, Sequence, Tuple, TypeVar, Union

import numpy as np
import torch as tr

from ._implementation import augment_and_mix

_T1 = TypeVar("_T1")
_T2 = TypeVar("_T2", np.ndarray, tr.Tensor, float)

__all__ = ["AugMix", "Fork", "augment_and_mix"]


def _check_non_neg_int(item, name):
    if not isinstance(item, (int, Integral)) or item < 0:
        raise ValueError(f"`{name}`` must be a non-negative integer. Got {item}")


class AugMix(tr.nn.Module):
    __doc__ = augment_and_mix.__doc__

    def __init__(
        self,
        process_fn: Callable[[_T1], _T2],
        augmentations: Sequence[Callable[[_T1], _T1]],
        *,
        num_aug_chains: int = 3,
        aug_chain_depth: Union[int, Tuple[int, int]] = (1, 4),
        beta_params: Union[float, Tuple[float, float]] = (1.0, 1.0),
        dirichlet_params: Union[float, Sequence[float]] = 1.0,
        augmentation_choice_probs: Optional[Sequence[float]] = None,
    ):
        super().__init__()
        _check_non_neg_int(num_aug_chains, "num_aug_chains")

        if isinstance(aug_chain_depth, Sequence):
            assert len(aug_chain_depth) == 2, aug_chain_depth
            _check_non_neg_int(aug_chain_depth[0], name="aug_chain_depth[0]")
            _check_non_neg_int(aug_chain_depth[1], name="aug_chain_depth[1]")
            assert aug_chain_depth[0] <= aug_chain_depth[1], aug_chain_depth

        if isinstance(dirichlet_params, Sequence):
            assert len(dirichlet_params) == num_aug_chains

        self.process_fn = process_fn
        self.augmentations = augmentations
        self.num_aug_chains = num_aug_chains
        self.aug_chain_depth = aug_chain_depth
        self.beta_params = beta_params
        self.dirichlet_params = dirichlet_params
        self.augmentations = augmentations
        self.augmentation_choice_probs = augmentation_choice_probs

        if augmentation_choice_probs is not None and len(
            augmentation_choice_probs
        ) != len(augmentations):
            raise ValueError(
                f"`len(sample_probabilities)` ({len(augmentation_choice_probs)}) must match `len(augmentations)` ({len(augmentations)})"
            )

    def forward(self, datum):
        return augment_and_mix(
            datum=datum,
            process_fn=self.process_fn,
            augmentations=self.augmentations,
            num_aug_chains=self.num_aug_chains,
            aug_chain_depth=self.aug_chain_depth,
            beta_params=self.beta_params,
            dirichlet_params=self.dirichlet_params,
            augmentation_choice_probs=self.augmentation_choice_probs,
        )

    def __repr__(self) -> str:
        format_string = self.__class__.__name__ + "(\naugmentations="
        for t in self.augmentations:
            format_string += "\n"
            format_string += f"    {t},"
        format_string += f"\nprocess_fn={self.process_fn},"
        format_string += f"\nnum_aug_chains={self.num_aug_chains},"
        format_string += f"\naug_chain_depth={self.aug_chain_depth},"
        format_string += f"\nbeta_params={self.beta_params},"
        format_string += f"\ndirichlet_params={self.dirichlet_params},"
        format_string += "\n)"

        return format_string


def _flat_repr(x) -> str:
    out = f"{x}".splitlines()
    if len(out) == 1:
        return out[0]
    else:
        return out[0] + "...)"


class Fork(tr.nn.Module):
    """
    Forks an input into an arbitrary number of transform-chains. This can
    be useful for doing consistency-loss workflows.

    Parameters
    ----------
    *forked_transforms: Callable[[Any], Any]
        One transform for each fork to create.

    Examples
    --------
    >>> from rai_toolbox.augmentations.augmix import Fork

    Here are some trivial examples:

    >>> two_fork = Fork(lambda x: x, lambda x: 2 * x)
    >>> two_fork(2)
    (2, 4)

    >>> three_fork = Fork(lambda x: x, lambda x: 2 * x, lambda x: 0 * x)
    >>> three_fork(-1.0)
    (-1.0, -2.0, -0.0)

    Here is a simplified version of the triple-processing used by the AugMix
    paper's consistency loss. It anticipates a PIL image and produces a triplet.

    >>> from torchvision.transforms import ToTensor, RandomHorizontalFlip, RandomVerticalFlip
    >>> from rai_toolbox.augmentations.augmix import AugMix
    >>> augmix = AugMix(
    ...     augmentations=[RandomHorizontalFlip(), RandomVerticalFlip()],
    ...     process_fn=ToTensor(),
    ... )
    >>> Fork(augmix, augmix, ToTensor())
    Fork(
          - ToTensor() ->
    x --> - AugMix(...) ->
          - AugMix(...) ->
    )
    """

    def __init__(self, *forked_transforms: Callable[[Any], Any]):
        super().__init__()

        if not forked_transforms:
            raise ValueError("At least one transform must be passed")

        if not all(callable(t) for t in forked_transforms):
            raise TypeError(
                f"All forked transforms must be callable, got: {forked_transforms}"
            )

        self.forked_transforms = forked_transforms

    def forward(self, x) -> Tuple[Any, ...]:
        return tuple(f(x) for f in self.forked_transforms)

    def __repr__(self) -> str:
        out = "Fork(\n"

        num_forks = len(self.forked_transforms)

        for n, f in enumerate(self.forked_transforms):
            if num_forks // 2 == n:
                if num_forks % 2 == 1:
                    out += f"x --> - {_flat_repr(f)} ->\n"
                else:
                    out += "x -->\n"
                    out += f"      - {_flat_repr(f)} ->\n"
            else:
                out += f"      - {_flat_repr(f)} ->\n"

        return out + ")"



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import os
import subprocess
import sys
from pathlib import Path
from time import sleep
from typing import Any, Callable, TypeVar

import numpy as np
from hydra.core.hydra_config import HydraConfig
from hydra_zen import load_from_yaml
from omegaconf.errors import ConfigAttributeError
from pytorch_lightning import Trainer
from pytorch_lightning.trainer.states import TrainerFn
from torch import distributed

from .._compatibility import PL_VERSION, Version

R = TypeVar("R")


def _setup_environment() -> None:
    if distributed.is_initialized():
        distributed.destroy_process_group()


def _teardown() -> None:
    # Remove PL environments so next multirun starts fresh
    envs = (
        "LOCAL_RANK",
        "NODE_RANK",
        "WORLD_SIZE",
        "MASTER_ADDR",
        "MASTER_PORT",
        "PL_GLOBAL_SEED",
    )

    for name in envs:
        os.environ.pop(name, None)


def _subprocess_call(local_rank: int, testing: bool) -> None:
    env_copy = os.environ.copy()
    env_copy["LOCAL_RANK"] = f"{local_rank}"
    # CWD is the Hydra working directory
    cwd = os.getcwd()
    os_cwd = (
        f'"{cwd}"'  # this is needed to handle characters like `=` in the directory name
    )

    command = [
        sys.executable,
        "-m",
        "rai_toolbox.mushin.lightning._pl_main",
    ]
    hydra_cfg = HydraConfig.get()

    hydra_output = (
        os.path.join(cwd, hydra_cfg.output_subdir)
        if hydra_cfg.output_subdir is not None
        else cwd
    )

    # Validate that minimal configuration requirements
    config = Path(hydra_output) / "config.yaml"
    assert config.exists()
    cfg = load_from_yaml(config)
    if "trainer" not in cfg or "module" not in cfg:
        raise ConfigAttributeError(
            "Missing configurations `trainer` and `module` are required for use with HydraDDP.  See documentation for further details."
        )

    # create the command for CLI
    command += ["-cp", hydra_output, "-cn", "config.yaml"]

    # Set flag to run Trainer.fit or Trainer.test in `_pl_main.py`
    command += ["++pl_testing=" + ("false" if not testing else "true")]

    # Set flag for local rank
    command += [f"++pl_local_rank={local_rank}"]

    command += [
        f"hydra.run.dir={os_cwd}",
        f"hydra.output_subdir=.pl_hydra_rank_{local_rank}",
        f"hydra.job.name={hydra_cfg.job.name}",
    ]
    subprocess.Popen(command, env=env_copy, cwd=cwd)


if PL_VERSION >= Version(1, 6, 0):
    from pytorch_lightning.strategies.ddp import DDPStrategy
    from pytorch_lightning.strategies.launchers.subprocess_script import (
        _SubprocessScriptLauncher,
    )

    class HydraDDP(DDPStrategy):  # type: ignore
        """DDP Strategy that supports Hydra run and multirun jobs.

        This strategy assumes a PyTorch Lightning `Trainer.fit` or `Trainer.test` has been configured
        to execute via Hydra.  It requires that Hydra saves a `config.yaml` in the current working directory with the following keys/properties set::

           ├── Config
           │    ├── trainer: A `pytorch_lightning.Trainer` configuration
           │    ├── module: A `pytorch_lightning.LightningModule` configuration
           │    ├── datamodule: [OPTIONAL] A `pytorch_lightning.LightningDataModule` configuration

        This strategy will launch a child subprocesses for additional GPU beyond the first using the following base command::

           python -m rai_toolbox.mushin.lightning._pl_main -cp <path to config.yaml> -cn config.yaml

        Examples
        --------

        First define a Hydra configuration using hydra-zen:

        >>> import pytorch_lightning as pl
        ... from hydra_zen import builds, make_config,
        ... from rai_toolbox.mushin import HydraDDP
        ... from rai_toolbox.mushin.testing.lightning import SimpleLightningModule
        ...
        ... TrainerConfig = builds(
        ...     pl.Trainer,
        ...     accelerator="auto",
        ...     gpus=2,
        ...     max_epochs=1,
        ...     fast_dev_run=True,
        ...     strategy=builds(HydraDDP),
        ...     populate_full_signature=True
        ... )
        ...
        ... ModuleConfig = builds(SimpleLightningModule)
        ...
        ... Config = make_config(
        ...     trainer=TrainerConfig,
        ...     module=ModuleConfig
        ... )

        Next, define a task function to execute the Hydra job:

        >>> from hydra_zen import instantiate
        >>> def task_function(cfg):
        ...     obj = instantiate(cfg)
        ...     obj.trainer.fit(obj.module)

        Launch the Hydra+Lightning DDP job

        >>> from hydra_zen import launch
        >>> job = launch(Config, task_function)

        ``HydraDDP`` also supports ``LightningDataModule`` configuration.

        >>> DataModuleConfig = ... # A LightningDataModule config
        >>> Config = make_config(
        ...     trainer=TrainerConfig,
        ...     module=ModuleConfig
        ...     datamodule=DataModuleconfig
        ... )

        Next define a task function to execute the Hydra job:

        >>> from hydra_zen import instantiate
        >>> def task_function(cfg):
        ...     obj = instantiate(cfg)
        ...     obj.trainer.fit(obj.module, datamodule=obj.datamodule)

        Launch the Hydra+Lightning DDP job:

        >>> from hydra_zen import launch
        >>> job = launch(Config, task_function)
        """

        def setup_environment(self) -> None:
            _setup_environment()
            super().setup_environment()

        def _configure_launcher(self) -> None:
            if self.cluster_environment is None:  # pragma: no cover
                raise TypeError("HydraDDP.cluster_environment is None")

            if not self.cluster_environment.creates_processes_externally:
                self._launcher = _HydraDDPLauncher(
                    self.cluster_environment, self.num_processes, self.num_nodes
                )
                self._rank_0_will_call_children_scripts = True

        def teardown(self) -> None:
            """Performs additional teardown steps for PL to allow for Hydra multirun jobs."""
            super().teardown()
            _teardown()

    class _HydraDDPLauncher(_SubprocessScriptLauncher):
        @property
        def is_interactive_compatible(self) -> bool:  # pragma: no cover
            return True

        def launch(
            self,
            function: Callable[..., R],
            *args: Any,
            trainer: Trainer,
            **kwargs: Any,
        ) -> R:
            """Creates new processes, then calls the given function.

            Parameters
            ----------
            function : Callable[[...], ReturnType]
                A callback function to execute after all processes have been created.
                It is up to the implementation of this function to synchronize the processes, e.g., with barriers.

            *args : Any
                Optional positional arguments to be passed to the given function.

            trainer : pytorch_lightning.Trainer
                Optional reference to the pytorch_lightning.Trainer`.

            **kwargs : Any
                Optional keyword arguments to be passed to the given function.

            Returns
            -------
            ReturnType
            """
            del trainer  # unused
            if not self.cluster_environment.creates_processes_externally:
                testing = function.__name__ == "_test_impl"
                self._call_children_scripts(testing=testing)

            return function(*args, **kwargs)

        def _call_children_scripts(self, testing: bool):
            # bookkeeping of spawned processes
            self._check_can_spawn_children()

            # DDP Environment variables
            os.environ["MASTER_ADDR"] = self.cluster_environment.main_address
            os.environ["MASTER_PORT"] = str(self.cluster_environment.main_port)

            # allow the user to pass the node rank
            os.environ["NODE_RANK"] = str(self.cluster_environment.node_rank())
            os.environ["LOCAL_RANK"] = str(self.cluster_environment.local_rank())
            os.environ["WORLD_SIZE"] = f"{self.num_processes * self.num_nodes}"

            for local_rank in range(1, self.num_processes):
                _subprocess_call(local_rank, testing)

                # starting all processes at once can cause issues
                # with dataloaders delay between 1-10 seconds
                delay = np.random.uniform(1, 5, 1)[0]
                sleep(delay)

else:  # pragma: no cover
    from pytorch_lightning.plugins.training_type.ddp import DDPPlugin

    class HydraDDP(DDPPlugin):
        """DDP Strategy that supports Hydra run and multirun jobs.

        This strategy assumes a PyTorch Lightning `Trainer.fit` or `Trainer.test` has been configured
        to execute via Hydra.  It requires that Hydra saves a `config.yaml` in the current working directory with the following keys/properties set::

           ├── Config
           │    ├── trainer: A `pytorch_lightning.Trainer` configuration
           │    ├── module: A `pytorch_lightning.LightningModule` configuration
           │    ├── datamodule: [OPTIONAL] A `pytorch_lightning.LightningDataModule` configuration

        This strategy will launch a child subprocesses for additional GPU beyond the first using the following base command::

           python -m rai_toolbox.mushin.lightning._pl_main -cp <path to config.yaml> -cn config.yaml

        Examples
        --------

        First define a Hydra configuration using hydra-zen:

        >>> import pytorch_lightning as pl
        ... from hydra_zen import builds, make_config,
        ... from rai_toolbox.mushin import HydraDDP
        ... from rai_toolbox.mushin.testing.lightning import SimpleLightningModule
        ...
        ... TrainerConfig = builds(
        ...     pl.Trainer,
        ...     accelerator="auto",
        ...     gpus=2,
        ...     max_epochs=1,
        ...     fast_dev_run=True,
        ...     strategy=builds(HydraDDP),
        ...     populate_full_signature=True
        ... )
        ...
        ... ModuleConfig = builds(SimpleLightningModule)
        ...
        ... Config = make_config(
        ...     trainer=TrainerConfig,
        ...     module=ModuleConfig
        ... )

        Next define a task function to execute the Hydra job:

        >>> from hydra_zen import instantiate
        >>> def task_function(cfg):
        ...     obj = instantiate(cfg)
        ...     obj.trainer.fit(obj.module)

        Launch the Hydra+Lightning DDP job:

        >>> from hydra_zen import launch
        >>> job = launch(Config, task_function)

        ``HydraDDP`` also supports ``LightningDataModule`` configuration.

        >>> DataModuleConfig = ... # A LightningDataModule config
        >>> Config = make_config(
        ...     trainer=TrainerConfig,
        ...     module=ModuleConfig
        ...     datamodule=DataModuleconfig
        ... )

        Next, define a task function to execute the Hydra job:

        >>> from hydra_zen import instantiate
        >>> def task_function(cfg):
        ...     obj = instantiate(cfg)
        ...     obj.trainer.fit(obj.module, datamodule=obj.datamodule)

        Launch the Hydra+Lightning DDP job:

        >>> from hydra_zen import launch
        >>> job = launch(Config, task_function)
        """

        def setup_environment(self) -> None:
            _setup_environment()
            super().setup_environment()

        def _call_children_scripts(self):
            if self.lightning_module is None:  # pragma: no cover
                raise TypeError("HydraDDP.lightning_module is None")

            if self.lightning_module.trainer is None:  # pragma: no cover
                raise TypeError("HydraDDP.lightning_module.trainer is None")

            if self.cluster_environment is None:  # pragma: no cover
                raise TypeError("HydraDDP.cluster_environment is None")

            # bookkeeping of spawned processes
            self._check_can_spawn_children()  # type: ignore

            # DDP Environment variables
            os.environ["MASTER_ADDR"] = self.cluster_environment.master_address()  # type: ignore
            os.environ["MASTER_PORT"] = str(self.cluster_environment.master_port())  # type: ignore

            # allow the user to pass the node rank
            os.environ["NODE_RANK"] = str(self.cluster_environment.node_rank())
            os.environ["LOCAL_RANK"] = str(self.cluster_environment.local_rank())
            os.environ["WORLD_SIZE"] = f"{self.num_processes * self.num_nodes}"

            self.interactive_ddp_procs = []
            for local_rank in range(1, self.num_processes):
                testing = self.lightning_module.trainer.state.fn == TrainerFn.TESTING
                _subprocess_call(local_rank, testing=testing)

                # starting all processes at once can cause issues
                # with dataloaders delay between 1-10 seconds
                delay = np.random.uniform(1, 5, 1)[0]
                sleep(delay)

            self._rank_0_has_called_call_children_scripts = True

        def teardown(self) -> None:
            """Performs additional teardown steps for PL to allow for Hydra multirun jobs."""
            super().teardown()
            _teardown()
